from datetime import date, timedelta
from decimal import Decimal
import logging

from django.conf import settings as django_settings
from django.db import connections, DatabaseError
from django.utils import timezone
import pyzabbix
from requests.exceptions import RequestException

from waldur_core.core.utils import datetime_to_timestamp, timestamp_to_datetime, QuietSession
from waldur_core.structure import ServiceBackend, ServiceBackendError, log_backend_action
from waldur_core.structure.utils import update_pulled_fields

from . import models, utils


logger = logging.getLogger(__name__)
sms_settings = getattr(django_settings, 'WALDUR_ZABBIX', {}).get('SMS_SETTINGS', {})


class ZabbixLogsFilter(logging.Filter):
    def filter(self, record):
        # Mute useless Zabbix log concerning JSON-RPC server endpoint.
        if record.getMessage().startswith('JSON-RPC Server Endpoint'):
            return False

        return super(ZabbixLogsFilter, self).filter(record)


pyzabbix.logger.addFilter(ZabbixLogsFilter())


class ZabbixBackendError(ServiceBackendError):
    pass


class ZabbixBackend(ServiceBackend):

    DEFAULTS = {
        'host_group_name': 'waldur',
        'templates_names': [],
        'database_parameters': {
            'host': 'localhost',
            'port': '3306',
            'name': 'zabbix',
            'user': 'admin',
            'password': '',
        },
        'interface_parameters': {
            'dns': '',
            'ip': '0.0.0.0',  # nosec
            'main': 1,
            'port': '10050',
            'type': 1,
            'useip': 1,
        },
        'sms_email_from': sms_settings.get('SMS_EMAIL_FROM'),
        'sms_email_rcpt': sms_settings.get('SMS_EMAIL_RCPT'),
    }

    TREND_DELAY_SECONDS = 60 * 60  # One hour
    HISTORY_DELAY_SECONDS = 15 * 60

    def __init__(self, settings):
        self.settings = settings

    @property
    def host_group_name(self):
        return self.settings.get_option('host_group_name')

    @property
    def templates_names(self):
        return self.settings.get_option('templates_names')

    @property
    def interface_parameters(self):
        return self.settings.get_option('interface_parameters')

    @property
    def database_parameters(self):
        return self.settings.get_option('database_parameters')

    @property
    def api(self):
        if not hasattr(self, '_api'):
            self._api = self._get_api(self.settings.backend_url,
                                      self.settings.username,
                                      self.settings.password)
        return self._api

    def ping(self, raise_exception=False):
        try:
            self.api.api_version()
        except Exception as e:
            if raise_exception:
                raise ZabbixBackendError(e)
            return False
        else:
            return True

    def pull_service_properties(self):
        self._get_or_create_group_id(self.host_group_name)
        self.pull_templates()
        self.pull_user_groups()
        self.pull_users()
        for name in self.templates_names:
            if not models.Template.objects.filter(name=name).exists():
                raise ZabbixBackendError('Cannot find template with name "%s".' % name)
        if not self.interface_parameters:
            raise ZabbixBackendError('Interface parameters should not be empty.')

    @log_backend_action()
    def create_host(self, host):
        interface_parameters = host.interface_parameters or self.interface_parameters
        host_group_name = host.host_group_name or self.host_group_name

        templates_ids = [t.backend_id for t in host.templates.all()]
        group_id, _ = self._get_or_create_group_id(host_group_name)

        zabbix_host_id, created = self._get_or_create_host_id(
            host_name=host.name,
            visible_name=host.visible_name,
            group_id=group_id,
            templates_ids=templates_ids,
            interface_parameters=interface_parameters,
            status=host.status,
        )

        if not created:
            logger.warning('Host with name "%s" already exists', host.name)

        host.interface_parameters = interface_parameters
        host.host_group_name = host_group_name
        host.backend_id = zabbix_host_id
        host.save()

    @log_backend_action()
    def update_host(self, host):
        try:
            group_id, _ = self._get_or_create_group_id(host.host_group_name)
            self.api.host.update({
                'hostid': host.backend_id,
                'host': host.name,
                'name': host.visible_name,
                'group_id': group_id,
                'templates': [{'templateid': t.backend_id} for t in host.templates.all()],
                'status': host.status,
            })
        except (pyzabbix.ZabbixAPIException, RequestException) as e:
            raise ZabbixBackendError(e)

    @log_backend_action()
    def delete_host(self, host):
        try:
            self.api.host.delete(host.backend_id)
        except (pyzabbix.ZabbixAPIException, RequestException) as e:
            raise ZabbixBackendError(e)

    @log_backend_action()
    def create_itservice(self, itservice):
        name = itservice.name or self._get_service_name(itservice.host.scope.backend_id)

        try:
            creation_kwargs = {
                'name': name,
                'algorithm': itservice.algorithm,
                'showsla': self._get_showsla(itservice.algorithm),
                'sortorder': itservice.sort_order,
                'goodsla': str(itservice.agreed_sla),
                'triggerid': None,
            }
            if itservice.trigger and itservice.host:
                creation_kwargs['triggerid'] = self._get_trigger_id(itservice.host.backend_id, itservice.trigger.name)

            data = self.api.service.create(creation_kwargs)
            service_id = data['serviceids'][0]
        except (pyzabbix.ZabbixAPIException, RequestException) as e:
            raise ZabbixBackendError(e)
        except (IndexError, KeyError) as e:
            raise ZabbixBackendError('ITService create request returned unexpected response: %s', data)
        else:
            itservice.backend_id = service_id
            itservice.name = name

            data = self.get_itservice(itservice.backend_id)
            itservice.backend_trigger_id = data['triggerid']
            itservice.save(update_fields=['backend_id', 'name', 'backend_trigger_id'])

    @log_backend_action()
    def delete_itservice(self, itservice):
        try:
            self.api.service.delete(itservice.backend_id)
        except (pyzabbix.ZabbixAPIException, RequestException) as e:
            raise ZabbixBackendError(e)

    def pull_templates(self):
        """ Update existing Waldur templates and their items """
        logger.debug('About to pull zabbix templates from backend.')
        try:
            zabbix_templates = self.api.template.get(
                output=['name', 'templateid'],
                selectTriggers=['description', 'triggerid', 'priority'],
                selectItems=['itemid', 'name', 'key_', 'value_type', 'units', 'history', 'delay'],
                selectTemplates=['templateid'],
            )
        except (pyzabbix.ZabbixAPIException, RequestException) as e:
            raise ZabbixBackendError('Cannot pull templates. Exception: %s' % e)

        # Delete stale templates
        zabbix_templates_ids = set([t['templateid'] for t in zabbix_templates])
        models.Template.objects.filter(settings=self.settings).exclude(backend_id__in=zabbix_templates_ids).delete()

        # Update or create zabbix templates
        for zabbix_template in zabbix_templates:
            nc_template, created = models.Template.objects.get_or_create(
                backend_id=zabbix_template['templateid'],
                settings=self.settings,
                defaults={'name': zabbix_template['name']})
            if not created and nc_template.name != zabbix_template['name']:
                nc_template.name = zabbix_template['name']
                nc_template.save()

            # Delete stale triggers
            zabbix_triggers_ids = set([i['triggerid'] for i in zabbix_template['triggers']])
            nc_template.triggers.exclude(backend_id__in=zabbix_triggers_ids).delete()

            # Update or create triggers
            for zabbix_trigger in zabbix_template['triggers']:
                nc_template.triggers.update_or_create(
                    backend_id=zabbix_trigger['triggerid'],
                    priority=int(zabbix_trigger['priority']),  # according to Zabbix model it must always be integer
                    settings=nc_template.settings,
                    defaults={'name': zabbix_trigger['description']})

            # Delete stale items
            zabbix_items_ids = set([i['itemid'] for i in zabbix_template['items']])
            nc_template.items.exclude(backend_id__in=zabbix_items_ids).delete()

            # Update or create zabbix items
            for zabbix_item in zabbix_template['items']:
                defaults = {
                    'name': zabbix_item['name'],
                    'key': zabbix_item['key_'],
                    'value_type': int(zabbix_item['value_type']),
                    'units': zabbix_item['units'],
                    'history': utils.parse_time(zabbix_item['history']),
                    'delay': utils.parse_time(zabbix_item['delay'])
                }

                nc_item, created = nc_template.items.get_or_create(
                    backend_id=zabbix_item['itemid'], defaults=defaults)
                if not created:
                    update_fields = []
                    for (name, value) in defaults.items():
                        if getattr(nc_item, name) != value:
                            setattr(nc_item, name, value)
                            update_fields.append(name)
                    if update_fields:
                        nc_item.save(update_fields=update_fields)

        # Initialize templates children
        for zabbix_template in zabbix_templates:
            if not zabbix_template['templates']:
                continue
            nc_template = models.Template.objects.get(settings=self.settings, backend_id=zabbix_template['templateid'])
            children_ids = [t['templateid'] for t in zabbix_template['templates']]
            children = models.Template.objects.filter(settings=self.settings, backend_id__in=children_ids)
            nc_template.children.add(*children)

        logger.info('Successfully pulled Zabbix templates for settings %s', self.settings)

    def get_item_last_value(self, host_id, key, **kwargs):
        try:
            items = self.api.item.get(hostids=host_id, filter={'key_': 'application.status'}, output=['lastvalue'])
        except pyzabbix.ZabbixAPIException as e:
            raise ZabbixBackendError('Cannot get zabbix items. Exception: %s' % e)
        try:
            return items[0]['lastvalue']
        except IndexError:
            raise ZabbixBackendError('Cannot find item with key "%s" for host with id %s' % (key, host_id))

    # XXX: This method should be rewrited - we need to pull only IT services that were connected to hosts.
    def pull_itservices(self):
        """
        Update IT services
        """
        logger.debug('About to pull Zabbix IT services')

        try:
            zabbix_services = self.api.service.get(output='extend')
        except pyzabbix.ZabbixAPIException as e:
            raise ZabbixBackendError('Cannot pull IT services. Exception: %s' % e)

        # Delete stale services
        zabbix_services_ids = set(i['serviceid'] for i in zabbix_services)
        models.ITService.objects.filter(settings=self.settings).exclude(backend_id__in=zabbix_services_ids).delete()

        triggers_map = self._get_triggers_map(zabbix_services)

        for zabbix_service in zabbix_services:
            triggerid = zabbix_service['triggerid']
            nc_triggerid = triggers_map.get(triggerid)

            defaults = {
                'name': zabbix_service['name'],
                'algorithm': int(zabbix_service['algorithm']),
                'sort_order': int(zabbix_service['sortorder']),
                'agreed_sla': Decimal(zabbix_service['goodsla']),
                'backend_trigger_id': triggerid,
                'trigger_id': nc_triggerid
            }
            nc_service, created = models.ITService.objects.get_or_create(
                settings=self.settings,
                backend_id=zabbix_service['serviceid'],
                defaults=defaults
            )
            if not created:
                update_fields = []
                for (name, value) in defaults.items():
                    if getattr(nc_service, name) != value:
                        setattr(nc_service, name, value)
                        update_fields.append(name)
                if update_fields:
                    nc_service.save(update_fields=update_fields)

        logger.info('Successfully pulled Zabbix IT services for settings %s.', self.settings)

    def pull_user_groups(self):
        logger.debug('About to pull Zabbix user groups.')

        try:
            zabbix_user_groups = self.api.usergroup.get()
        except (pyzabbix.ZabbixAPIException, RequestException) as e:
            raise ZabbixBackendError('Cannot pull user groups. Exception: %s' % e)
        # Delete stale
        zabbix_user_group_ids = set(i['usrgrpid'] for i in zabbix_user_groups)
        models.UserGroup.objects.filter(settings=self.settings).exclude(backend_id__in=zabbix_user_group_ids).delete()
        # Update or create
        for zabbix_user_group in zabbix_user_groups:
            models.UserGroup.objects.update_or_create(
                backend_id=zabbix_user_group['usrgrpid'],
                settings=self.settings,
                defaults={'name': zabbix_user_group['name']})
        logger.info('Successfully pulled Zabbix user groups for settings %s.', self.settings)

    def pull_users(self):
        logger.debug('About to pull Zabbix users.')

        try:
            zabbix_users = self.api.user.get(selectUsrgrps=['name', 'usrgrpid'], output='extend')
        except (pyzabbix.ZabbixAPIException, RequestException) as e:
            raise ZabbixBackendError('Cannot pull users. Exception: %s' % e)
        # Delete stale
        zabbix_users_ids = set(i['userid'] for i in zabbix_users)
        models.User.objects.filter(settings=self.settings).exclude(backend_id__in=zabbix_users_ids).delete()
        # Update or create
        for zabbix_user in zabbix_users:
            groups = []
            for zabbix_user_group in zabbix_user['usrgrps']:
                group, _ = models.UserGroup.objects.get_or_create(
                    backend_id=zabbix_user_group['usrgrpid'],
                    settings=self.settings,
                    defaults={'name': zabbix_user_group['name']})
                groups.append(group)
            user, _ = models.User.objects.update_or_create(
                backend_id=zabbix_user['userid'],
                settings=self.settings,
                defaults={
                    'name': zabbix_user['name'],
                    'alias': zabbix_user['alias'],
                    'surname': zabbix_user['surname'],
                    'type': int(zabbix_user['type']),
                    'state': models.User.States.OK,
                })
            user.groups.add(*groups)

        logger.info('Successfully pulled Zabbix users for settings %s.', self.settings)

    @log_backend_action()
    def create_user(self, user):
        try:
            zabbix_user = self.api.user.create(
                name=user.name,
                surname=user.surname,
                alias=user.alias,
                type=user.type,
                passwd=user.password,
                usrgrps=[{'usrgrpid': group.backend_id for group in user.groups.all()}])
        except (pyzabbix.ZabbixAPIException, RequestException) as e:
            raise ZabbixBackendError(e)
        user.backend_id = zabbix_user['userids'][0]
        user.save(update_fields=['backend_id'])

    @log_backend_action()
    def update_user(self, user):
        try:
            self.api.user.update(
                userid=user.backend_id,
                name=user.name,
                surname=user.surname,
                alias=user.alias,
                type=user.type,
                passwd=user.password,
                usrgrps=[{'usrgrpid': group.backend_id for group in user.groups.all()}])
        except (pyzabbix.ZabbixAPIException, RequestException) as e:
            raise ZabbixBackendError(e)

    @log_backend_action()
    def delete_user(self, user):
        try:
            self.api.user.delete(user.backend_id)
        except (pyzabbix.ZabbixAPIException, RequestException) as e:
            raise ZabbixBackendError(e)

    # XXX: This method is hotfix for user group permissions management. We
    #      should create models for permissions. NC-1564.
    @log_backend_action()
    def add_permission_to_user_group(self, user_group, host_group_name, permission_id):
        try:
            host_group_id, _ = self._get_or_create_group_id(host_group_name)
            self.api.usergroup.update(
                usrgrpid=user_group.backend_id,
                rights=[{'id': host_group_id, 'permission': permission_id}]
            )
        except (pyzabbix.ZabbixAPIException, RequestException) as e:
            raise ZabbixBackendError(e)

    def _get_triggers_map(self, zabbix_services):
        """
        Return map of Zabbix trigger ID to Waldur trigger ID
        """
        trigger_ids = self._map_keys(zabbix_services, 'triggerid')

        zabbix_triggers = []
        try:
            zabbix_triggers = self.api.trigger.get(triggerids=trigger_ids, output=['triggerid', 'templateid'])
        except (pyzabbix.ZabbixAPIException, RequestException) as e:
            logger.exception('Unable to fetch Zabbix triggers')
            raise ZabbixBackendError(e)

        template_ids = self._map_keys(zabbix_triggers, 'templateid')
        nc_triggers = models.Trigger.objects.filter(
            settings=self.settings,
            backend_id__in=template_ids).only('pk')
        nc_triggers_map = {trigger.backend_id: trigger.pk for trigger in list(nc_triggers)}

        return {trigger['triggerid']: nc_triggers_map.get(trigger['templateid'])
                for trigger in zabbix_triggers}

    def _map_keys(self, items, key):
        return list(set(item[key] for item in items))

    def _get_or_create_group_id(self, group_name):
        try:
            exists = self.api.hostgroup.get(filter={'name': group_name})
            if not exists:
                group_id = self.api.hostgroup.create({'name': group_name})['groupids'][0]
                return group_id, True
            else:
                return self.api.hostgroup.get(filter={'name': group_name})[0]['groupid'], False
        except (pyzabbix.ZabbixAPIException, IndexError, KeyError) as e:
            raise ZabbixBackendError('Cannot get or create group with name "%s". Exception: %s' % (group_name, e))

    def _get_or_create_host_id(self, host_name, visible_name, group_id, templates_ids, interface_parameters, status):
        """ Create Zabbix host with given parameters.

        Return (<host_id>, <is_created>) tuple as result.
        """
        host_id = self._get_host_id(host_name)
        if host_id:
            return host_id, False
        host_id = self._create_host(host_name, visible_name, group_id, templates_ids,
                                    interface_parameters, status)
        return host_id, True

    def _get_host_id(self, host_name):
        try:
            host = self.api.host.get(filter={'host': host_name}, output='hostid')
            # If host with given name does not exist, empty list is returned
            if host:
                return host[0]['hostid']
        except (pyzabbix.ZabbixAPIException, RequestException) as e:
            logger.error('Cannot get host id by name: %s. Exception: %s', host_name, str(e))
            raise ZabbixBackendError(e)

    def _create_host(self, host_name, visible_name, group_id, templates_ids, interface_parameters, status):
        host_parameters = {
            "host": host_name,
            "name": visible_name,
            "interfaces": [interface_parameters],
            "groups": [{"groupid": group_id}],
            "templates": [{'templateid': template_id} for template_id in templates_ids],
            "status": status,
        }

        try:
            host = self.api.host.create(host_parameters)
            return host['hostids'][0]
        except (pyzabbix.ZabbixAPIException, RequestException) as e:
            logger.error('Cannot create host with parameters: %s. Exception: %s',
                         host_parameters, str(e))
            raise ZabbixBackendError(e)

    def _get_showsla(self, algorithm):
        if algorithm == models.ITService.Algorithm.ANY:
            return 1
        elif algorithm == models.ITService.Algorithm.ALL:
            return 1
        else:
            return 0

    def _get_service_name(self, backend_id):
        return 'Availability of %s' % backend_id

    def _get_trigger_id(self, host_id, description):
        """
        Find trigger ID by host ID and trigger description
        """
        try:
            data = self.api.trigger.get(
                filter={'description': description},
                hostids=host_id,
                output=['triggerid'])
            return data[0]['triggerid']
        except (pyzabbix.ZabbixAPIException, RequestException, IndexError, KeyError) as e:
            logger.exception('No trigger for host %s and description %s', host_id, description)
            raise ZabbixBackendError(e)

    def get_sla(self, service_id, start_time, end_time):
        try:
            data = self.api.service.getsla(
                filter={'serviceids': service_id},
                intervals={'from': start_time, 'to': end_time}
            )
            return data[service_id]['sla'][0]['sla']
        except (pyzabbix.ZabbixAPIException, RequestException, IndexError, KeyError) as e:
            message = 'Can not get Zabbix IT service SLA value for service with ID %s. Exception: %s'
            raise ZabbixBackendError(message % (service_id, e))

    def get_itservice(self, service_id):
        try:
            response = self.api.service.get(filter={'serviceid': service_id}, output='extend')
        except (pyzabbix.ZabbixAPIException, RequestException) as e:
            raise ZabbixBackendError('Can not get status of Zabbix IT service with ID %s. Exception: %s',
                                     service_id, e)
        if len(response) != 1:
            raise ZabbixBackendError('Zabbix IT service with ID %s is not found. '
                                     'Response is %s', service_id, response)
        return response[0]

    def get_sla_range(self, serviceid):
        """
        Execute query to Zabbix DB to get minimum and maximum clock for service's alarm.
        Returns minimum and maximum dates.
        """
        query = 'SELECT min(clock), max(clock) FROM service_alarms WHERE serviceid = %s'
        cursor = self._execute_query(query, [serviceid])
        min_timestamp, max_timestamp = cursor.fetchone()
        return date.fromtimestamp(int(min_timestamp)), date.fromtimestamp(int(max_timestamp))

    def get_trigger_events(self, trigger_id, start_time, end_time):
        try:
            event_data = self.api.event.get(
                output='extend',
                objectids=trigger_id,
                time_from=start_time,
                time_till=end_time,
                sortfield=["clock"],
                sortorder="ASC")
        except (pyzabbix.ZabbixAPIException, RequestException) as e:
            message = 'Can not get events for trigger with ID %s. Exception: %s'
            raise ZabbixBackendError(message % (trigger_id, e))
        else:
            return [{'timestamp': e['clock'], 'value': e['value']} for e in event_data]

    def _get_api(self, backend_url, username, password):
        unsafe_session = QuietSession()
        unsafe_session.verify = False

        api = pyzabbix.ZabbixAPI(server=backend_url, session=unsafe_session)
        api.login(username, password)
        return api

    def get_item_stats(self, hostid, item, points):
        if item.value_type == models.Item.ValueTypes.FLOAT:
            history_table = 'history'
            trend_table = 'trends'
        elif item.value_type == models.Item.ValueTypes.INTEGER:
            # Integer value
            history_table = 'history_uint'
            trend_table = 'trends_uint'
        else:
            raise ZabbixBackendError('Cannot get statistics for non-numerical item %s' % item.key)

        history_retention_days = item.history
        history_delay_seconds = item.delay or self.HISTORY_DELAY_SECONDS
        trend_delay_seconds = self.TREND_DELAY_SECONDS

        trends_start_date = datetime_to_timestamp(timezone.now() - timedelta(days=history_retention_days))

        points = points[::-1]
        history_cursor = self._get_history(
            item.key, hostid, history_table, points[-1] - history_delay_seconds, points[0])
        trends_cursor = self._get_history(
            item.key, hostid, trend_table, points[-1] - trend_delay_seconds, points[0])

        values = []
        if points[0] > trends_start_date:
            next_value = history_cursor.fetchone()
        else:
            next_value = trends_cursor.fetchone()

        for end, start in zip(points[:-1], points[1:]):
            if start > trends_start_date:
                interval = history_delay_seconds
            else:
                interval = trend_delay_seconds

            value = None
            while True:
                if next_value is None:
                    break
                time, value = next_value
                if item.is_byte():
                    value = self.b2mb(value)

                if time <= end:
                    if end - time < interval or time > start:
                        break
                else:
                    if start > trends_start_date:
                        next_value = history_cursor.fetchone()
                    else:
                        next_value = trends_cursor.fetchone()

            values.append(value)
        return values[::-1]

    def get_items_aggregated_values(self, host, items, start_timestamp, end_timestamp, method='MAX'):
        """
        Get aggregate values for host items.

        Output format:
            {
                <item1.name>: <aggregated value>,
                <item2.name>: <aggregated value>,
                ...
            }
        """
        int_items = [item for item in items if item.value_type == models.Item.ValueTypes.INTEGER]
        float_items = [item for item in items if item.value_type == models.Item.ValueTypes.FLOAT]

        # Get aggregated data from DB
        db_data = tuple()
        default_kwargs = {
            'hostid': host.backend_id,
            'start_timestamp': start_timestamp,
            'end_timestamp': end_timestamp,
            'method': method,
        }
        # XXX: We need to get values from table "trends" if end_timestamp < item.history.
        if int_items:
            cursor = self._get_aggregated_values(
                item_keys=[item.key for item in int_items],
                table='history_uint',
                **default_kwargs
            )
            db_data += cursor.fetchall()
        if float_items:
            cursor = self._get_aggregated_values(
                item_keys=[item.key for item in float_items],
                table='history',
                **default_kwargs
            )
            db_data += cursor.fetchall()

        # Prepare data - convert B to MB if needed
        items_keys = {item.key: item for item in items}
        aggregated_values = {}
        for key, value in db_data:
            item = items_keys[key]
            aggregated_values[key] = self.b2mb(value) if item.is_byte() else value
        return aggregated_values

    def b2mb(self, value):
        return value / 1024 / 1024

    def _get_history(self, item_key, hostid, table, start_timestamp, end_timestamp):
        """
        Execute query to zabbix db to get item values from history
        """
        query = (
            'SELECT clock time, %(value_path)s value '
            'FROM %(table)s history, items '
            'WHERE history.itemid = items.itemid '
            'AND items.key_ = "%(item_key)s" '
            'AND items.hostid = %(hostid)s '
            'AND clock > %(start_timestamp)s '
            'AND clock < %(end_timestamp)s '
            'ORDER BY clock DESC'
        )
        parameters = {
            'table': table,
            'item_key': item_key,
            'hostid': hostid,
            'start_timestamp': start_timestamp,
            'end_timestamp': end_timestamp,
            'value_path': table.startswith('history') and 'value' or 'value_avg'
        }
        query = query % parameters

        return self._execute_query(query)

    def _get_aggregated_values(self, hostid, item_keys, start_timestamp, end_timestamp, table, method='MAX'):
        """
        Execute query to zabbix DB to get item aggregated historical value.
        """
        # XXX: This query is really slow with a lot of item_keys, need to speed up it with index.
        query = (
            'SELECT items.key_, %(method)s(value) '
            'FROM hosts, items, %(table)s history '
            'WHERE items.hostid = %(hostid)s AND hosts.hostid = %(hostid)s AND history.itemid = items.itemid '
            'AND items.key_ IN (%(item_keys)s) '
            'AND clock >= %(start_timestamp)s '
            'AND clock <= %(end_timestamp)s '
            'GROUP BY items.key_'
        )

        parameters = {
            'method': method,
            'table': table,
            'hostid': hostid,
            'item_keys': ', '.join(['"%s"' % key for key in item_keys]),
            'start_timestamp': start_timestamp,
            'end_timestamp': end_timestamp,
        }

        query = query % parameters
        return self._execute_query(query)

    def _get_db_connection(self, force=False):
        host = self.database_parameters['host']
        port = self.database_parameters['port']
        name = self.database_parameters['name']
        user = self.database_parameters['user']
        password = self.database_parameters['password']

        key = '/'.join([name, host, port])
        if key not in connections.databases or force:
            connections.databases[key] = {
                'ENGINE': 'django.db.backends.mysql',
                'NAME': name,
                'HOST': host,
                'PORT': port,
                'USER': user,
                'PASSWORD': password
            }
        return connections[key]

    def _execute_query(self, query, *args, **kwargs):
        logger.debug('Executing query %s to Zabbix' % query)
        try:
            cursor = self._get_db_connection().cursor()
            cursor.execute(query, *args, **kwargs)
            return cursor
        except DatabaseError as e:
            logger.exception('Can not execute query the Zabbix DB.')
            raise ZabbixBackendError(e)

    def import_host(self, host_backend_id, service_project_link=None, save=True):
        if save and not service_project_link:
            raise AttributeError('Cannot save imported host if SPL is not defined.')
        try:
            backend_host = self.api.host.get(
                filter={'hostid': host_backend_id}, selectGroups=True,
                output=['host', 'name', 'description', 'error', 'status', 'groups'])[0]
        except IndexError:
            raise ZabbixBackendError('Host with id %s does not exist at backend' % host_backend_id)
        except (pyzabbix.ZabbixAPIException, RequestException) as e:
            raise ZabbixBackendError(e)

        host = models.Host()
        host.name = backend_host['host']
        host.visible_name = backend_host['name']
        host.description = backend_host['description']
        host.error = backend_host['error']
        host.status = backend_host['status']
        host.backend_id = host_backend_id
        if backend_host.get('groups'):
            # Host groups list is serialized as in following example:
            # [{u'internal': u'0', u'flags': u'0', u'groupid': u'15', u'name': u'waldur'}]
            host.host_group_name = backend_host['groups'][0]['name']
        else:
            host.host_group_name = ''

        if save:
            host.service_project_link = service_project_link
            host.save()
            templates = self.get_host_templates(host)
            host.templates.add(*templates)
        return host

    def get_host_templates(self, host):
        try:
            backend_templates = self.api.template.get(hostids=[host.backend_id], output=['templateid'])
        except (pyzabbix.ZabbixAPIException, RequestException) as e:
            raise ZabbixBackendError(e)
        return models.Template.objects.filter(
            backend_id__in=[t['templateid'] for t in backend_templates],
            settings=host.service_project_link.service.settings)

    @log_backend_action()
    def pull_host(self, host):
        import_time = timezone.now()
        imported_host = self.import_host(host.backend_id, save=False)
        imported_host_templates = set(self.get_host_templates(host))

        host.refresh_from_db()
        if host.modified < import_time:
            update_fields = ('name', 'visible_name', 'description', 'error', 'status', 'host_group_name')
            update_pulled_fields(host, imported_host, update_fields)

        host_templates = set(host.templates.all())
        host.templates.remove(*(host_templates - imported_host_templates))
        host.templates.add(*(imported_host_templates - host_templates))

    def get_trigger_request(self, query):
        request = {}

        request['selectHosts'] = 1
        request['active'] = 1
        request['expandComment'] = 1
        request['expandDescription'] = 1
        request['expandExpression'] = 1

        if 'host_id' in query:
            request['hostids'] = query['host_id']

        if 'host_name' in query:
            request['host'] = query['host_name']

        if 'changed_after' in query:
            request['lastChangeSince'] = datetime_to_timestamp(query['changed_after'])

        if 'changed_before' in query:
            request['lastChangeTill'] = datetime_to_timestamp(query['changed_before'])

        if 'min_priority' in query:
            request['min_severity'] = query['min_priority']

        if 'priority' in query and len(query['priority']) > 0:
            request['filter'] = {'priority': list(query['priority'])}

        if 'acknowledge_status' in query:
            acknowledge_status = query['acknowledge_status']
            Status = models.Trigger.AcknowledgeStatus
            status_mapping = {
                Status.SOME_EVENTS_UNACKNOWLEDGED: 'withUnacknowledgedEvents',
                Status.LAST_EVENT_UNACKNOWLEDGED: 'withLastEventUnacknowledged',
                Status.ALL_EVENTS_ACKNOWLEDGED: 'withAcknowledgedEvents',
            }
            if acknowledge_status in status_mapping:
                key = status_mapping[acknowledge_status]
                request[key] = 1

        if 'value' in query:
            request.setdefault('filter', {})
            request['filter']['value'] = query['value']

        return request

    def get_trigger_status(self, query):
        request = self.get_trigger_request(query)
        get_events_count = query.get('include_events_count')
        get_trigger_hosts = query.get('include_trigger_hosts')
        try:
            backend_triggers = self.api.trigger.get(**request)

            backend_events = None
            if get_events_count:
                objectids = map(lambda t: t['triggerid'], backend_triggers)
                backend_events = self.api.event.get(objectids=objectids,
                                                    acknowledged=0,
                                                    countOutput=True,
                                                    groupCount=True,
                                                    value='1')  # 1 means that trigger has a problem
                # https://www.zabbix.com/documentation/3.4/manual/api/reference/event/object

            trigger_hosts = None
            if get_trigger_hosts:
                objectids = map(lambda t: t['triggerid'], backend_triggers)
                trigger_hosts = self.api.host.get(triggerids=objectids)

            triggers = []

            for trigger in backend_triggers:
                triggers.append(self._parse_trigger(trigger, backend_events, trigger_hosts))

            return triggers
        except (pyzabbix.ZabbixAPIException, RequestException) as e:
            logger.exception('Unable to fetch Zabbix triggers')
            raise ZabbixBackendError(e)

    def _parse_trigger(self, backend_trigger, backend_events=None, trigger_hosts=None):
        trigger = {}

        for field in django_settings.WALDUR_ZABBIX['TRIGGER_FIELDS']:
            trigger[field[0]] = backend_trigger[field[1]]

        trigger['changed'] = timestamp_to_datetime(backend_trigger['lastchange'])
        trigger['hosts'] = []

        for host in backend_trigger['hosts']:
            host_id = host['hostid']

            if trigger_hosts is not None:
                host_name = filter(lambda h: h['hostid'] == host_id, trigger_hosts)
                if host_name:
                    host_name = host_name[0]['host']
                else:
                    host_name = ''
            else:
                host_name = None

            trigger['hosts'].append({'id': host_id, 'name': host_name})

        trigger['event_count'] = None
        if backend_events is not None:
            events = filter(lambda e: e['objectid'] == trigger['backend_id'], backend_events)
            trigger['event_count'] = 0 if not events else events[0]['rowscount']

        return trigger

    def get_trigger_count(self, query):
        request = self.get_trigger_request(query)
        try:
            return int(self.api.trigger.get(countOutput=True, **request))
        except (pyzabbix.ZabbixAPIException, RequestException) as e:
            logger.exception('Unable to fetch Zabbix triggers')
            raise ZabbixBackendError(e)



from waldur_zabbix import views as zabbix_views

from . import filters, serializers


class LinkViewSet(zabbix_views.ZabbixServiceProjectLinkViewSet):
    serializer_class = serializers.LinkSerializer
    filter_backends = zabbix_views.ZabbixServiceProjectLinkViewSet.filter_backends + (
        filters.LinkFilterBackend,
    )



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-08-07 18:47
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('openstack_tenant', '0001_squashed_0046'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('core', '0001_squashed_0008'),
        ('playbook_jobs', '0003_playbook_image'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='job',
            name='project',
        ),
        migrations.AddField(
            model_name='job',
            name='service_project_link',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='openstack_tenant.OpenStackTenantServiceProjectLink'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='job',
            name='ssh_public_key',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='core.SshPublicKey'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='job',
            name='subnet',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='openstack_tenant.SubNet'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='job',
            name='user',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.CASCADE, related_name='+', to=settings.AUTH_USER_MODEL),
            preserve_default=False,
        ),
    ]



from . import SupportBackend


class BasicBackend(SupportBackend):
    def create_issue(self, issue):
        return

    def update_issue(self, issue):
        return

    def delete_issue(self, issue):
        return

    def create_comment(self, comment):
        return

    def update_comment(self, comment):
        return

    def delete_comment(self, comment):
        return

    def create_attachment(self, attachment):
        return

    def delete_attachment(self, attachment):
        return

    def get_users(self):
        return

    def pull_priorities(self):
        return



from django.apps import AppConfig
from django.db.models import signals
from django_fsm import signals as fsm_signals


class OpenStackTenantConfig(AppConfig):
    """ OpenStack is a toolkit for building private and public clouds.
        This application adds support for managing OpenStack tenant resources -
        instances, volumes and snapshots.
    """
    name = 'waldur_openstack.openstack_tenant'
    label = 'openstack_tenant'
    verbose_name = 'OpenStackTenant'
    service_name = 'OpenStackTenant'

    def ready(self):
        from waldur_core.quotas.fields import QuotaField, TotalQuotaField
        from waldur_core.quotas.models import Quota
        from waldur_core.structure.models import ServiceSettings, Project, Customer
        from waldur_core.structure import SupportedServices
        from waldur_openstack.openstack.models import Tenant

        from .backend import OpenStackTenantBackend
        from . import handlers, models

        SupportedServices.register_backend(OpenStackTenantBackend)

        # Initialize service settings quotas based on tenant.
        for quota in Tenant.get_quotas_fields():
            ServiceSettings.add_quota_field(
                name=quota.name,
                quota_field=QuotaField(
                    is_backend=True,
                    default_limit=quota.default_limit,
                    creation_condition=lambda service_settings:
                        service_settings.type == OpenStackTenantConfig.service_name
                )
            )

        signals.post_save.connect(
            handlers.sync_private_settings_quotas_with_tenant_quotas,
            sender=Quota,
            dispatch_uid='openstack_tenant.handlers.sync_private_settings_quotas_with_tenant_quotas',
        )

        Project.add_quota_field(
            name='os_cpu_count',
            quota_field=TotalQuotaField(
                target_models=[models.Instance],
                path_to_scope='service_project_link.project',
                target_field='cores',
            )
        )

        Project.add_quota_field(
            name='os_ram_size',
            quota_field=TotalQuotaField(
                target_models=[models.Instance],
                path_to_scope='service_project_link.project',
                target_field='ram',
            )
        )

        Project.add_quota_field(
            name='os_storage_size',
            quota_field=TotalQuotaField(
                target_models=[models.Volume, models.Snapshot],
                path_to_scope='service_project_link.project',
                target_field='size',
            )
        )

        Customer.add_quota_field(
            name='os_cpu_count',
            quota_field=TotalQuotaField(
                target_models=[models.Instance],
                path_to_scope='service_project_link.project.customer',
                target_field='cores',
            )
        )

        Customer.add_quota_field(
            name='os_ram_size',
            quota_field=TotalQuotaField(
                target_models=[models.Instance],
                path_to_scope='service_project_link.project.customer',
                target_field='ram',
            )
        )

        Customer.add_quota_field(
            name='os_storage_size',
            quota_field=TotalQuotaField(
                target_models=[models.Volume, models.Snapshot],
                path_to_scope='service_project_link.project.customer',
                target_field='size',
            )
        )

        for Resource in (models.Instance, models.Volume, models.Snapshot):
            name = Resource.__name__.lower()
            signals.post_save.connect(
                handlers.log_action,
                sender=Resource,
                dispatch_uid='openstack_tenant.handlers.log_%s_action' % name,
            )

        for handler in handlers.resource_handlers:
            model = handler.resource_model
            name = model.__name__.lower()

            fsm_signals.post_transition.connect(
                handler.create_handler,
                sender=model,
                dispatch_uid='openstack_tenant.handlers.create_%s' % name,
            )

            fsm_signals.post_transition.connect(
                handler.update_handler,
                sender=model,
                dispatch_uid='openstack_tenant.handlers.update_%s' % name,
            )

            signals.post_save.connect(
                handler.import_handler,
                sender=model,
                dispatch_uid='openstack_tenant.handlers.import_%s' % name,
            )

            signals.post_delete.connect(
                handler.delete_handler,
                sender=model,
                dispatch_uid='openstack_tenant.handlers.delete_%s' % name,
            )

        signals.post_save.connect(
            handlers.log_backup_schedule_creation,
            sender=models.BackupSchedule,
            dispatch_uid='openstack_tenant.handlers.log_backup_schedule_creation',
        )

        signals.post_save.connect(
            handlers.log_backup_schedule_action,
            sender=models.BackupSchedule,
            dispatch_uid='openstack_tenant.handlers.log_backup_schedule_action',
        )

        signals.pre_delete.connect(
            handlers.log_backup_schedule_deletion,
            sender=models.BackupSchedule,
            dispatch_uid='openstack_tenant.handlers.log_backup_schedule_deletion',
        )

        signals.post_save.connect(
            handlers.log_snapshot_schedule_creation,
            sender=models.SnapshotSchedule,
            dispatch_uid='openstack_tenant.handlers.log_snapshot_schedule_creation',
        )

        signals.post_save.connect(
            handlers.log_snapshot_schedule_action,
            sender=models.SnapshotSchedule,
            dispatch_uid='openstack_tenant.handlers.log_snapshot_schedule_action',
        )

        signals.pre_delete.connect(
            handlers.log_snapshot_schedule_deletion,
            sender=models.SnapshotSchedule,
            dispatch_uid='openstack_tenant.handlers.log_snapshot_schedule_deletion',
        )

        signals.post_save.connect(
            handlers.update_service_settings_credentials,
            sender=Tenant,
            dispatch_uid='openstack_tenant.handlers.update_service_settings_credentials',
        )

        signals.post_save.connect(
            handlers.update_service_settings,
            sender=Tenant,
            dispatch_uid='openstack_tenant.handlers.update_service_settings',
        )

        signals.m2m_changed.connect(
            handlers.sync_certificates_between_openstack_service_with_openstacktenant_service,
            sender=ServiceSettings.certifications.through,
            dispatch_uid='openstack_tenant.handlers.'
                         'sync_certificates_between_openstack_service_with_openstacktenant_service',
        )

        signals.post_save.connect(
            handlers.copy_certifications_from_openstack_service_to_openstacktenant_service,
            sender=ServiceSettings,
            dispatch_uid='openstack_tenant.handlers.'
                         'copy_certifications_from_openstack_service_to_openstacktenant_service',
        )

        signals.post_save.connect(
            handlers.copy_flavor_exclude_regex_to_openstacktenant_service_settings,
            sender=ServiceSettings,
            dispatch_uid='openstack_tenant.handlers.'
                         'copy_flavor_exclude_regex_to_openstacktenant_service_settings',
        )

        signals.post_save.connect(
            handlers.create_service_from_tenant,
            sender=Tenant,
            dispatch_uid='openstack_tenant.handlers.create_service_from_tenant',
        )

        signals.post_save.connect(
            handlers.sync_price_list_item_for_flavor,
            sender=models.Flavor,
            dispatch_uid='openstack_tenant.handlers.sync_price_list_item_for_flavor',
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2018-12-20 11:41
from django.db import migrations, models
import django.utils.timezone
import model_utils.fields
import waldur_core.core.fields


class Migration(migrations.Migration):

    dependencies = [
        ('logging', '0004_json_field'),
    ]

    operations = [
        migrations.CreateModel(
            name='Report',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('file', models.FileField(upload_to='logging_reports')),
                ('file_size', models.PositiveIntegerField(null=True)),
                ('state', models.CharField(choices=[('pending', 'Pending'), ('done', 'Done'), ('erred', 'Erred')], default='pending', max_length=10)),
                ('error_message', models.TextField(blank=True)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-03-05 22:40
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_slurm', '0005_add_deposit'),
    ]

    operations = [
        migrations.AddField(
            model_name='allocationusage',
            name='deposit_usage',
            field=models.DecimalField(decimal_places=2, default=0, max_digits=8),
        ),
    ]



from waldur_core.core import WaldurExtension


class MarketplaceAzureExtension(WaldurExtension):

    @staticmethod
    def django_app():
        return 'waldur_mastermind.marketplace_azure'

    @staticmethod
    def is_assembly():
        return True



from django.db import models
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.utils.translation import ugettext_lazy as _
from model_utils import FieldTracker

from waldur_core.core import models as core_models
from waldur_core.logging import models as logging_models
from waldur_core.structure import models as structure_models
from waldur_mastermind.invoices import models as invoices_models
from waldur_mastermind.invoices import utils as invoices_utils

from . import exceptions, managers


class PriceEstimate(logging_models.AlertThresholdMixin, core_models.UuidMixin, models.Model):
    content_type = models.ForeignKey(on_delete=models.CASCADE, to=ContentType, null=True, related_name='+')
    object_id = models.PositiveIntegerField(null=True)
    scope = GenericForeignKey('content_type', 'object_id')

    objects = managers.PriceEstimateManager('scope')
    tracker = FieldTracker()

    total = models.FloatField(default=0, help_text=_('Predicted price for scope for current month.'))
    limit = models.FloatField(default=-1, help_text=_('Price limit of a scope object in current month. '
                                                      '-1 means no limit.'))

    def is_over_threshold(self):  # For AlertThresholdMixin
        return self.total > self.threshold

    def validate_limit(self):
        if self.limit != -1 and self.total > self.limit:
            raise exceptions.PriceEstimateLimitExceeded(self)

    @classmethod
    def get_estimated_models(cls):
        return structure_models.Project, structure_models.Customer

    def _get_sum(self, year, month, field):
        items = invoices_models.InvoiceItem.objects.filter(
            invoice__year=year,
            invoice__month=month
        )
        if self.content_type.model_class() == structure_models.Project:
            items = items.filter(project__uuid=self.scope.uuid.hex)
        elif self.content_type.model_class() == structure_models.Customer:
            items = items.filter(invoice__customer=self.scope)
        return sum(getattr(item, field) for item in items)

    def get_total(self, year, month, current=False):
        return self._get_sum(year, month, current and 'price_current' or 'price')

    def get_tax(self, year, month, current=False):
        return self._get_sum(year, month, current and 'tax_current' or 'tax')

    def update_total(self):
        current_year = invoices_utils.get_current_year()
        current_month = invoices_utils.get_current_month()
        self.total = self.get_total(current_year, current_month)



from __future__ import unicode_literals

import logging

from celery import shared_task
from django.contrib import auth
from django.contrib.contenttypes.models import ContentType
from rest_framework import status
from rest_framework.reverse import reverse

from waldur_core.core import tasks as core_tasks, utils as core_utils
from waldur_core.structure.signals import resource_imported
from waldur_mastermind.common import utils as common_utils
from waldur_openstack.openstack_tenant import models as openstack_tenant_models
from waldur_openstack.openstack_tenant.views import InstanceViewSet

from waldur_rancher.utils import SyncUser

from . import models, exceptions, signals, utils, views

logger = logging.getLogger(__name__)


class CreateNodeTask(core_tasks.Task):
    def execute(self, instance, user_id):
        node = instance
        content_type = ContentType.objects.get_for_model(openstack_tenant_models.Instance)
        flavor = node.initial_data['flavor']
        system_volume_size = node.initial_data['system_volume_size']
        system_volume_type = node.initial_data.get('system_volume_type')
        data_volumes = node.initial_data.get('data_volumes', [])
        image = node.initial_data['image']
        subnet = node.initial_data['subnet']
        group = node.initial_data['group']
        tenant_spl = node.initial_data['tenant_service_project_link']
        user = auth.get_user_model().objects.get(pk=user_id)

        post_data = {
            'name': node.name,
            'flavor': reverse('openstacktenant-flavor-detail', kwargs={'uuid': flavor}),
            'image': reverse('openstacktenant-image-detail', kwargs={'uuid': image}),
            'service_project_link': reverse('openstacktenant-spl-detail', kwargs={'pk': tenant_spl}),
            'system_volume_size': system_volume_size,
            'system_volume_type': system_volume_type and reverse('openstacktenant-volume-type-detail', kwargs={'uuid': system_volume_type}),
            'data_volumes': [{
                'size': volume['size'],
                'volume_type': volume.get('volume_type') and reverse('openstacktenant-volume-type-detail', kwargs={'uuid': volume.get('volume_type')}),
            } for volume in data_volumes],
            'security_groups': [{'url': reverse('openstacktenant-sgp-detail', kwargs={'uuid': group})}],
            'internal_ips_set': [
                {
                    'subnet': reverse('openstacktenant-subnet-detail', kwargs={'uuid': subnet})
                }
            ],
            'user_data': utils.format_node_cloud_config(node),
        }
        view = InstanceViewSet.as_view({'post': 'create'})
        response = common_utils.create_request(view, user, post_data)

        if response.status_code != status.HTTP_201_CREATED:
            raise exceptions.RancherException(response.data)

        instance_uuid = response.data['uuid']
        instance = openstack_tenant_models.Instance.objects.get(uuid=instance_uuid)
        node.content_type = content_type
        node.object_id = instance.id
        node.state = models.Node.States.CREATING
        node.save()

        resource_imported.send(
            sender=instance.__class__,
            instance=instance,
        )

    @classmethod
    def get_description(cls, instance, *args, **kwargs):
        return 'Create nodes for k8s cluster "%s".' % instance


class DeleteNodeTask(core_tasks.Task):
    def execute(self, instance, user_id):
        node = instance
        user = auth.get_user_model().objects.get(pk=user_id)

        if node.instance:
            view = InstanceViewSet.as_view({'delete': 'destroy'})
            response = common_utils.delete_request(view, user, uuid=node.instance.uuid.hex)

            if response.status_code != status.HTTP_202_ACCEPTED:
                raise exceptions.RancherException(response.data)
        else:
            backend = node.cluster.get_backend()
            backend.delete_node(node)


@shared_task
def update_nodes(cluster_id):
    cluster = models.Cluster.objects.get(id=cluster_id)
    backend = cluster.get_backend()

    if cluster.node_set.filter(backend_id='').exists():
        backend_nodes = backend.get_cluster_nodes(cluster.backend_id)

        for backend_node in backend_nodes:
            if cluster.node_set.filter(name=backend_node['name']).exists():
                node = cluster.node_set.get(name=backend_node['name'])
                node.backend_id = backend_node['backend_id']
                node.save()

    has_changes = False

    for node in cluster.node_set.exclude(backend_id=''):
        old_state = node.state
        backend.update_node_details(node)
        node.refresh_from_db()

        if old_state != node.state:
            has_changes = True

    if has_changes:
        signals.node_states_have_been_updated.send(
            sender=models.Cluster,
            instance=cluster,
        )


@shared_task(name='waldur_rancher.update_clusters_nodes')
def update_clusters_nodes():
    for cluster in models.Cluster.objects.exclude(backend_id=''):
        update_nodes.delay(cluster.id)


class PollRuntimeStateNodeTask(core_tasks.Task):
    max_retries = 1200
    default_retry_delay = 30

    @classmethod
    def get_description(cls, node, *args, **kwargs):
        node = core_utils.deserialize_instance(node)
        return 'Poll node "%s"' % node.name

    def execute(self, node):
        update_nodes(node.cluster_id)
        node.refresh_from_db()

        if not node.backend_id:
            self.retry()

        return node


@shared_task(name='waldur_rancher.notify_create_user')
def notify_create_user(id, password, url):
    user = models.RancherUser.objects.get(id=id)
    email = user.user.email

    context = {
        'rancher_url': url,
        'user': user,
        'password': password,
    }

    core_utils.broadcast_mail('rancher', 'notification_create_user', context, [email])


@shared_task(name='waldur_rancher.sync_users')
def sync_users():
    SyncUser.run()


class DeleteClusterNodesTask(core_tasks.Task):
    def execute(self, instance, user_id):
        cluster = instance
        view = views.NodeViewSet.as_view({'delete': 'destroy'})
        user = auth.get_user_model().objects.get(pk=user_id)

        for node in cluster.node_set.all():
            response = common_utils.delete_request(view, user, uuid=node.uuid.hex)

            if response.status_code != status.HTTP_202_ACCEPTED:
                node.error_message = 'Instance deleting\'s an error: %s.' % response.data
                node.set_erred()
                node.save()

    @classmethod
    def get_description(cls, instance, *args, **kwargs):
        return 'Delete nodes for k8s cluster "%s".' % instance



from django.conf import settings
from django.core.validators import MinValueValidator, MaxValueValidator
from django.db import models
from django.utils.translation import ugettext_lazy as _
from model_utils import FieldTracker

from waldur_core.structure import models as structure_models
from waldur_slurm import utils


def get_batch_service(service_settings):
    batch_service = service_settings.options.get('batch_service')
    if batch_service not in ('SLURM', 'MOAB'):
        batch_service = 'SLURM'
    return batch_service


class SlurmService(structure_models.Service):
    projects = models.ManyToManyField(structure_models.Project,
                                      related_name='+', through='SlurmServiceProjectLink')

    class Meta:
        unique_together = ('customer', 'settings')
        verbose_name = _('SLURM provider')
        verbose_name_plural = _('SLURM providers')

    @classmethod
    def get_url_name(cls):
        return 'slurm'


class SlurmServiceProjectLink(structure_models.ServiceProjectLink):
    service = models.ForeignKey(on_delete=models.CASCADE, to=SlurmService)

    class Meta(structure_models.ServiceProjectLink.Meta):
        verbose_name = _('SLURM provider project link')
        verbose_name_plural = _('SLURM provider project links')

    @classmethod
    def get_url_name(cls):
        return 'slurm-spl'


class Allocation(structure_models.NewResource):
    service_project_link = models.ForeignKey(
        SlurmServiceProjectLink, related_name='allocations', on_delete=models.PROTECT)
    is_active = models.BooleanField(default=True)
    tracker = FieldTracker()

    cpu_limit = models.BigIntegerField(default=-1)
    cpu_usage = models.BigIntegerField(default=0)

    gpu_limit = models.BigIntegerField(default=-1)
    gpu_usage = models.BigIntegerField(default=0)

    ram_limit = models.BigIntegerField(default=-1)
    ram_usage = models.BigIntegerField(default=0)

    deposit_limit = models.DecimalField(max_digits=6, decimal_places=0, default=-1)
    deposit_usage = models.DecimalField(max_digits=8, decimal_places=2, default=0)

    @classmethod
    def get_url_name(cls):
        return 'slurm-allocation'

    def usage_changed(self):
        return any(self.tracker.has_changed(field) for field in utils.FIELD_NAMES)

    @classmethod
    def get_backend_fields(cls):
        return super(Allocation, cls).get_backend_fields() + (
            'cpu_usage', 'gpu_usage', 'ram_usage', 'deposit_usage'
        )

    @property
    def batch_service(self):
        return get_batch_service(self.service_project_link.service.settings)


class AllocationUsage(models.Model):
    class Permissions:
        customer_path = 'allocation__service_project_link__project__customer'
        project_path = 'allocation__service_project_link__project'
        service_path = 'allocation__service_project_link__service'

    class Meta:
        ordering = ['allocation']

    allocation = models.ForeignKey(on_delete=models.CASCADE, to=Allocation)
    username = models.CharField(max_length=32)
    user = models.OneToOneField(settings.AUTH_USER_MODEL, blank=True, null=True, on_delete=models.CASCADE)

    year = models.PositiveSmallIntegerField()
    month = models.PositiveSmallIntegerField(validators=[MinValueValidator(1), MaxValueValidator(12)])

    cpu_usage = models.BigIntegerField(default=0)
    ram_usage = models.BigIntegerField(default=0)
    gpu_usage = models.BigIntegerField(default=0)
    deposit_usage = models.DecimalField(max_digits=8, decimal_places=2, default=0)



from django.contrib import admin
from django.utils.translation import ugettext_lazy as _

from waldur_core.structure import admin as structure_admin

from .models import SlurmService, SlurmServiceProjectLink, Allocation


def get_allocation_count(self, scope):
    return scope.quotas.get(name='nc_allocation_count').usage


get_allocation_count.short_description = _('Allocation count')

for cls in (structure_admin.CustomerAdmin, structure_admin.ProjectAdmin):
    cls.get_allocation_count = get_allocation_count
    cls.list_display += ('get_allocation_count',)

admin.site.register(SlurmService, structure_admin.ServiceAdmin)
admin.site.register(SlurmServiceProjectLink, structure_admin.ServiceProjectLinkAdmin)
admin.site.register(Allocation, structure_admin.ResourceAdmin)



import datetime

from django.utils import timezone

from waldur_core.core import utils as core_utils
from waldur_mastermind.invoices import registrators
from waldur_mastermind.support import models as support_models
from waldur_mastermind.marketplace import models as marketplace_models

from .utils import is_request_based, component_usage_register

OrderTypes = marketplace_models.OrderItem.Types


def add_new_offering_to_invoice(sender, instance, created=False, **kwargs):
    if created:
        return

    order_item = instance

    if order_item.tracker.has_changed('state') and \
            order_item.state == marketplace_models.OrderItem.States.DONE and \
            order_item.type == marketplace_models.OrderItem.Types.CREATE and \
            order_item.resource and \
            order_item.resource.scope and \
            isinstance(order_item.resource.scope, support_models.Offering) and \
            is_request_based(order_item.resource.scope):
        offering = support_models.Offering.objects.get(pk=order_item.resource.scope.pk)
        registrators.RegistrationManager.register(offering, timezone.now(), order_type=OrderTypes.CREATE)


def terminate_invoice_when_offering_deleted(sender, instance, **kwargs):
    if not is_request_based(instance):
        return

    request_based_offering = support_models.Offering.objects.get(pk=instance.pk)
    registrators.RegistrationManager.terminate(request_based_offering, timezone.now())


def terminate_invoice_when_offering_cancelled(sender, instance, created=False, **kwargs):
    if created:
        return

    if not is_request_based(instance):
        return

    if instance.tracker.has_changed('state') and (instance.state == support_models.Offering.States.TERMINATED):
        request_based_offering = support_models.Offering.objects.get(pk=instance.pk)
        registrators.RegistrationManager.terminate(request_based_offering, timezone.now())


def switch_plan_resource(sender, instance, created=False, **kwargs):
    if created:
        return

    resource = instance
    support_offering = resource.scope

    if not isinstance(support_offering, support_models.Offering):
        return

    if not resource.tracker.has_changed('plan_id'):
        return

    registrators.RegistrationManager.terminate(support_offering, timezone.now())

    if resource.plan.scope:
        support_offering.plan = resource.plan.scope
        support_offering.save(update_fields=['plan'])

    support_offering.unit_price = instance.plan.unit_price
    support_offering.unit = instance.plan.unit
    support_offering.save()

    registrators.RegistrationManager.register(support_offering, timezone.now(), order_type=OrderTypes.UPDATE)


def update_invoice_on_offering_deletion(sender, instance, **kwargs):
    state = instance.state

    if is_request_based(instance):
        return

    if state == support_models.Offering.States.TERMINATED:
        # no need to terminate offering item if it was already terminated before.
        return

    registrators.RegistrationManager.terminate(instance, timezone.now())


def add_new_offering_details_to_invoice(sender, instance, created=False, **kwargs):
    state = instance.state

    if is_request_based(instance):
        return

    if (state == support_models.Offering.States.OK and
            support_models.Offering.States.REQUESTED == instance.tracker.previous('state')):
        registrators.RegistrationManager.register(instance, timezone.now(), order_type=OrderTypes.CREATE)
    if (state == support_models.Offering.States.TERMINATED and
            support_models.Offering.States.OK == instance.tracker.previous('state')):
        registrators.RegistrationManager.terminate(instance, timezone.now())


def add_component_usage(sender, instance, created=False, **kwargs):
    component_usage = instance

    if not created and not component_usage.tracker.has_changed('usage'):
            return

    if not isinstance(component_usage.resource.scope, support_models.Offering):
        return

    component_usage_register(component_usage)


def create_recurring_usage_if_invoice_has_been_created(sender, instance, created=False, **kwargs):
    if not created:
        return

    invoice = instance

    now = timezone.now()
    prev_month = (now.replace(day=1) - datetime.timedelta(days=1)).date()
    prev_month_start = prev_month.replace(day=1)
    usages = marketplace_models.ComponentUsage.objects.filter(
        resource__project__customer=invoice.customer,
        recurring=True,
        billing_period__gte=prev_month_start
    )

    if not usages:
        return

    for usage in usages:
        marketplace_models.ComponentUsage.objects.create(
            resource=usage.resource,
            component=usage.component,
            usage=usage.usage,
            description=usage.description,
            date=now,
            plan_period=usage.plan_period,
            recurring=usage.recurring,
            billing_period=core_utils.month_start(now),
        )



from django.apps import AppConfig
from django.db.models import signals


class QuotasConfig(AppConfig):
    """
    Quotas - objects resource limits and their usage.
    Quotas limits can be editable by users.
    """
    name = 'waldur_core.quotas'
    verbose_name = 'Quotas'

    def ready(self):
        from waldur_core.quotas import handlers, utils

        Quota = self.get_model('Quota')

        for index, model in enumerate(utils.get_models_with_quotas()):
            signals.post_save.connect(
                handlers.increase_global_quota,
                sender=model,
                dispatch_uid='waldur_core.quotas.handlers.increase_global_quota_%s_%s' % (model.__name__, index)
            )

            signals.post_delete.connect(
                handlers.decrease_global_quota,
                sender=model,
                dispatch_uid='waldur_core.quotas.handlers.decrease_global_quota_%s_%s' % (model.__name__, index)
            )

        signals.post_migrate.connect(
            handlers.create_global_quotas,
            dispatch_uid="waldur_core.quotas.handlers.create_global_quotas",
        )

        # new quotas
        from waldur_core.quotas import fields

        for model_index, model in enumerate(utils.get_models_with_quotas()):
            # quota initialization
            signals.post_save.connect(
                handlers.init_quotas,
                sender=model,
                dispatch_uid='waldur_core.quotas.init_quotas_%s_%s' % (model.__name__, model_index)
            )

            # Counter quota signals
            # How it works:
            # Each counter quota field has list of target models. Change of target model should increase or decrease
            # counter quota. So we connect generated handler to each of target models.
            for counter_field in model.get_quotas_fields(field_class=fields.CounterQuotaField):
                self.register_counter_field_signals(model, counter_field)

        # Aggregator quotas signals
        signals.post_save.connect(
            handlers.handle_aggregated_quotas,
            sender=Quota,
            dispatch_uid='waldur_core.quotas.handle_aggregated_quotas_post_save',
        )

        signals.pre_delete.connect(
            handlers.handle_aggregated_quotas,
            sender=Quota,
            dispatch_uid='waldur_core.quotas.handle_aggregated_quotas_pre_delete',
        )

    @staticmethod
    def register_counter_field_signals(model, counter_field):
        from waldur_core.quotas import handlers

        for target_model_index, target_model in enumerate(counter_field.target_models):
            signals.post_save.connect(
                handlers.count_quota_handler_factory(counter_field),
                sender=target_model,
                weak=False,  # saves handler from garbage collector
                dispatch_uid='waldur_core.quotas.increase_counter_quota_%s_%s_%s_%s_%s' % (
                    model.__name__, model._meta.app_label, counter_field.name, target_model.__name__, target_model_index)
            )

            signals.post_delete.connect(
                handlers.count_quota_handler_factory(counter_field),
                sender=target_model,
                weak=False,  # saves handler from garbage collector
                dispatch_uid='waldur_core.quotas.decrease_counter_quota_%s_%s_%s_%s_%s' % (
                    model.__name__, model._meta.app_label, counter_field.name, target_model.__name__, target_model_index)
            )



from django.conf.urls import url

from . import views


def register_in(router):
    pass


urlpatterns = [
    url(r'^api/ansible-estimator/$', views.AnsibleEstimatorView.as_view(), name='ansible-estimator'),
]



from collections import OrderedDict

default_app_config = 'waldur_openstack.openstack_tenant.apps.OpenStackTenantConfig'


class PriceItemTypes:
    FLAVOR = 'flavor'
    STORAGE = 'storage'
    LICENSE_APPLICATION = 'license-application'
    LICENSE_OS = 'license-os'
    SUPPORT = 'support'

    CHOICES = (
        (FLAVOR, 'flavor'),
        (STORAGE, 'storage'),
        (LICENSE_APPLICATION, 'license-application'),
        (LICENSE_OS, 'license-os'),
        (SUPPORT, 'support'),
    )


class OsTypes:
    CENTOS6 = 'centos6'
    CENTOS7 = 'centos7'
    UBUNTU = 'ubuntu'
    RHEL6 = 'rhel6'
    RHEL7 = 'rhel7'
    FREEBSD = 'freebsd'
    WINDOWS = 'windows'
    OTHER = 'other'

    CHOICES = (
        (CENTOS6, 'Centos 6'),
        (CENTOS7, 'Centos 7'),
        (UBUNTU, 'Ubuntu'),
        (RHEL6, 'RedHat 6'),
        (RHEL7, 'RedHat 7'),
        (FREEBSD, 'FreeBSD'),
        (WINDOWS, 'Windows'),
        (OTHER, 'Other'),
    )

    CATEGORIES = OrderedDict([
        ('Linux', (CENTOS6, CENTOS7, UBUNTU, RHEL6, RHEL7)),
        ('Windows', (WINDOWS,)),
        ('Other', (FREEBSD, OTHER)),
    ])


class ApplicationTypes:
    WORDPRESS = 'wordpress'
    POSTGRESQL = 'postgresql'
    ZIMBRA = 'zimbra'
    ZABBIX = 'zabbix'
    SUGAR = 'sugar'

    CHOICES = (
        (WORDPRESS, 'WordPress'),
        (POSTGRESQL, 'PostgreSQL'),
        (ZIMBRA, 'Zimbra'),
        (ZABBIX, 'Zabbix'),
        (SUGAR, 'Sugar'),
    )


class SupportTypes:
    BASIC = 'basic'
    PREMIUM = 'premium'
    ADVANCED = 'advanced'

    CHOICES = (
        (BASIC, 'Basic'),
        (PREMIUM, 'Premium'),
        (ADVANCED, 'Advanced'),
    )


class Types:
    PriceItems = PriceItemTypes
    Applications = ApplicationTypes
    Support = SupportTypes
    Os = OsTypes



from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import serializers

from waldur_core.core import serializers as core_serializers
from waldur_core.structure import models as structure_models
from waldur_zabbix.apps import ZabbixConfig


class LinkFilterBackend(DjangoFilterBackend):
    """
    This filter allows to filter Zabbix service project link by URL of virtual machine.
    Consider for example the following use case.

    There're two OpenStack virtual machines in the Waldur project.
    Zabbix server is installed on the first VM.
    Zabbix agent is to be installed on the second VM.
    Note, that both of them share the same OpenStack tenant.
    Therefore, they should be able to communicate directly, ie without proxy or virtual router.

    There's service settings for Zabbix provider in Waldur database.
    It is configured with scope field equal to the Zabbix server VM.
    Also, there are Zabbix service and Zabbix service project link configured for the project.

    By supplying URL of the OpenStack service project link to this filter backend,
    we should be able to get list of all Zabbix service project links
    which could be used as Zabbix monitoring in the same OpenStack tenant.
    """

    def filter_queryset(self, request, queryset, view):
        resource_url = request.query_params.get('resource')
        if resource_url:
            try:
                resource = self.get_resource_by_url(request, resource_url)
            except serializers.ValidationError:
                return queryset.none()

            link = resource.service_project_link
            siblings = resource._meta.model.objects.filter(
                service_project_link=link
            ).exclude(uuid=resource.uuid)
            if siblings.count() == 0:
                return queryset.none()

            service_settings = structure_models.ServiceSettings.objects.filter(
                type=ZabbixConfig.service_name,
                scope__in=siblings,
            )
            queryset = queryset.filter(project=link.project, service__settings__in=service_settings)
        return queryset

    def get_resource_by_url(self, request, resource_url):
        related_models = structure_models.VirtualMachine.get_all_models()
        field = core_serializers.GenericRelatedField(related_models=related_models)
        # Trick to set field context without serializer
        field._context = {'request': request}
        return field.to_internal_value(resource_url)



from . import views


def register_in(router):
    router.register(r'zabbix', views.ZabbixServiceViewSet, basename='zabbix')
    router.register(r'zabbix-hosts', views.HostViewSet, basename='zabbix-host')
    router.register(r'zabbix-itservices', views.ITServiceViewSet, basename='zabbix-itservice')
    router.register(r'zabbix-service-project-link', views.ZabbixServiceProjectLinkViewSet, basename='zabbix-spl')
    router.register(r'zabbix-templates', views.TemplateViewSet, basename='zabbix-template')
    router.register(r'zabbix-triggers', views.TriggerViewSet, basename='zabbix-trigger')
    router.register(r'zabbix-user-groups', views.UserGroupViewSet, basename='zabbix-user-group')
    router.register(r'zabbix-users', views.UserViewSet, basename='zabbix-user')



from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core import validators as django_validators
from django.db import transaction
from django.utils.translation import ugettext_lazy as _
from rest_framework import serializers

from waldur_core.core import serializers as core_serializers
from waldur_core.structure import models as structure_models
from waldur_core.structure import serializers as structure_serializers
from waldur_core.structure.models import VirtualMachine
from waldur_openstack.openstack_tenant import apps as openstack_tenant_apps
from waldur_openstack.openstack_tenant import models as openstack_tenant_models

from . import models, validators, exceptions, utils


class ServiceSerializer(core_serializers.ExtraFieldOptionsMixin,
                        structure_serializers.BaseServiceSerializer):

    SERVICE_ACCOUNT_FIELDS = {
        'backend_url': _('Rancher server URL'),
        'username': _('Rancher access key'),
        'password': _('Rancher secret key'),
    }

    SERVICE_ACCOUNT_EXTRA_FIELDS = {
        'base_image_name': _('Base image name'),
        'cloud_init_template': _('Cloud init template'),
    }

    class Meta(structure_serializers.BaseServiceSerializer.Meta):
        model = models.RancherService
        required_fields = ('backend_url', 'username', 'password', 'base_image_name')


class ServiceProjectLinkSerializer(structure_serializers.BaseServiceProjectLinkSerializer):
    class Meta(structure_serializers.BaseServiceProjectLinkSerializer.Meta):
        model = models.RancherServiceProjectLink
        extra_kwargs = {
            'service': {'lookup_field': 'uuid', 'view_name': 'vmware-detail'},
        }


class DataVolumeSerializer(structure_serializers.PermissionFieldFilteringMixin,
                           serializers.Serializer):
    size = serializers.IntegerField()
    volume_type = serializers.HyperlinkedRelatedField(
        view_name='openstacktenant-volume-type-detail',
        queryset=openstack_tenant_models.VolumeType.objects.all(),
        lookup_field='uuid',
        allow_null=True,
        required=False,
    )

    def get_fields(self):
        fields = super(DataVolumeSerializer, self).get_fields()
        fields['mount_point'] = serializers.ChoiceField(
            choices=settings.WALDUR_RANCHER['MOUNT_POINT_CHOICES']
        )
        return fields

    def get_filtered_field_names(self):
        return ['volume_type']

    def validate(self, attrs):
        size = attrs['size']
        mount_point = attrs['mount_point']
        min_size = settings.WALDUR_RANCHER['MOUNT_POINT_MIN_SIZE'][mount_point]
        if size < min_size * 1024:
            raise serializers.ValidationError('Volume %s capacity should be at least %s GB' %
                                              (mount_point, min_size))
        return attrs


class BaseNodeSerializer(structure_serializers.PermissionFieldFilteringMixin,
                         serializers.HyperlinkedModelSerializer):
    ROLE_CHOICES = ('controlplane', 'etcd', 'worker')
    subnet = serializers.HyperlinkedRelatedField(
        view_name='openstacktenant-subnet-detail',
        queryset=openstack_tenant_models.SubNet.objects.all(),
        lookup_field='uuid',
        allow_null=True,
        write_only=True,
    )
    flavor = serializers.HyperlinkedRelatedField(
        view_name='openstacktenant-flavor-detail',
        queryset=openstack_tenant_models.Flavor.objects.all(),
        lookup_field='uuid',
        allow_null=True,
        write_only=True,
        required=False,
    )
    system_volume_size = serializers.IntegerField(
        write_only=True,
        required=False,
        validators=[django_validators.MinValueValidator(
            lambda: settings.WALDUR_RANCHER['SYSTEM_VOLUME_MIN_SIZE'])],
    )
    system_volume_type = serializers.HyperlinkedRelatedField(
        view_name='openstacktenant-volume-type-detail',
        queryset=openstack_tenant_models.VolumeType.objects.all(),
        lookup_field='uuid',
        allow_null=True,
        required=False,
        write_only=True,
    )
    data_volumes = DataVolumeSerializer(many=True, write_only=True, required=False)
    memory = serializers.IntegerField(write_only=True, required=False)
    cpu = serializers.IntegerField(write_only=True, required=False)
    roles = serializers.MultipleChoiceField(choices=ROLE_CHOICES, write_only=True)

    class Meta(object):
        model = models.Node

    def get_filtered_field_names(self):
        return ('subnet', 'flavor', 'system_volume_type')


class NestedNodeSerializer(BaseNodeSerializer):
    instance = core_serializers.GenericRelatedField(
        related_models=VirtualMachine.get_all_models(),
        read_only=True
    )

    class Meta(BaseNodeSerializer.Meta):
        extra_kwargs = {
            'url': {'lookup_field': 'uuid', 'view_name': 'rancher-node-detail'},
            'cluster': {'lookup_field': 'uuid', 'view_name': 'rancher-cluster-detail'}
        }
        exclude = ('cluster', 'object_id', 'content_type', 'name')


class ClusterSerializer(structure_serializers.BaseResourceSerializer):
    service = serializers.HyperlinkedRelatedField(
        source='service_project_link.service',
        view_name='rancher-detail',
        read_only=True,
        lookup_field='uuid',
    )

    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='rancher-spl-detail',
        queryset=models.RancherServiceProjectLink.objects.all(),
        allow_null=True,
        required=False,
    )

    tenant_settings = serializers.HyperlinkedRelatedField(
        queryset=structure_models.ServiceSettings.objects.filter(
            type=openstack_tenant_apps.OpenStackTenantConfig.service_name),
        view_name='servicesettings-detail',
        lookup_field='uuid',
    )

    name = serializers.CharField(max_length=150, validators=[validators.ClusterNameValidator])
    nodes = NestedNodeSerializer(many=True, source='node_set')

    class Meta(structure_serializers.BaseResourceSerializer.Meta):
        model = models.Cluster
        fields = structure_serializers.BaseResourceSerializer.Meta.fields + (
            'node_command', 'nodes', 'tenant_settings', 'runtime_state',
        )
        read_only_fields = structure_serializers.BaseResourceSerializer.Meta.read_only_fields + (
            'node_command', 'runtime_state',
        )
        protected_fields = structure_serializers.BaseResourceSerializer.Meta.protected_fields + (
            'nodes', 'tenant_settings',
        )
        extra_kwargs = dict(
            cluster={
                'view_name': 'rancher-cluster-detail',
                'lookup_field': 'uuid',
            },
            **structure_serializers.BaseResourceSerializer.Meta.extra_kwargs
        )

    def validate(self, attrs):
        # Skip validation on update
        if self.instance:
            return attrs

        nodes = attrs.get('node_set')
        name = attrs.get('name')
        spl = attrs.get('service_project_link')
        tenant_settings = attrs.get('tenant_settings')
        utils.expand_added_nodes(nodes, spl, tenant_settings, name)
        return super(ClusterSerializer, self).validate(attrs)

    def validate_nodes(self, nodes):
        if len([node for node in nodes if 'etcd' in node['roles']]) not in [1, 3, 5]:
            raise serializers.ValidationError(
                'Total count of etcd nodes must be 1, 3 or 5. You have got %s nodes.' % len(nodes)
            )

        if not len([node for node in nodes if 'worker' in node['roles']]):
            raise serializers.ValidationError('Count of workers roles must be min 1.')

        if not len([node for node in nodes if 'controlplane' in node['roles']]):
            raise serializers.ValidationError('Count of controlplane nodes must be min 1.')

        return nodes


class NodeSerializer(serializers.HyperlinkedModelSerializer):
    instance = core_serializers.GenericRelatedField(
        related_models=VirtualMachine.get_all_models(),
        required=True,
    )
    resource_type = serializers.SerializerMethodField()
    state = serializers.ReadOnlyField(source='get_state_display')
    service_name = serializers.ReadOnlyField(source='service_project_link.service.settings.name')
    service_uuid = serializers.ReadOnlyField(source='service_project_link.service.uuid')
    project_uuid = serializers.ReadOnlyField(source='service_project_link.project.uuid')
    cluster_name = serializers.ReadOnlyField(source='cluster.name')
    cluster_uuid = serializers.ReadOnlyField(source='cluster.uuid')
    instance_name = serializers.ReadOnlyField(source='instance.name')
    instance_uuid = serializers.ReadOnlyField(source='instance.uuid')

    class Meta:
        model = models.Node
        fields = ('uuid', 'url', 'created', 'modified', 'name', 'backend_id', 'project_uuid',
                  'service_name', 'service_uuid', 'resource_type', 'state',
                  'cluster', 'cluster_name', 'cluster_uuid',
                  'instance', 'instance_name', 'instance_uuid',
                  'controlplane_role', 'etcd_role', 'worker_role', 'get_node_command', 'k8s_version',
                  'docker_version', 'cpu_allocated', 'cpu_total', 'ram_allocated', 'ram_total', 'pods_allocated',
                  'pods_total', 'labels', 'annotations', 'runtime_state')
        read_only_fields = ('backend_id', 'k8s_version', 'docker_version', 'cpu_allocated', 'cpu_total',
                            'ram_allocated', 'ram_total', 'pods_allocated', 'pods_total', 'labels', 'annotations',
                            'runtime_state')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid', 'view_name': 'rancher-node-detail'},
            'cluster': {'lookup_field': 'uuid', 'view_name': 'rancher-cluster-detail'}
        }

    def validate(self, attrs):
        instance = attrs.get('instance')

        if models.Node.objects.filter(
                object_id=instance.id,
                content_type=ContentType.objects.get_for_model(instance)
        ).exists():
            raise serializers.ValidationError({'instance': 'The selected instance is already in use.'})

        attrs['name'] = instance.name

        return super(NodeSerializer, self).validate(attrs)

    def get_resource_type(self, obj):
        return 'Rancher.Node'


class CreateNodeSerializer(BaseNodeSerializer):
    class Meta:
        model = models.Node
        fields = ('cluster', 'roles', 'system_volume_size', 'system_volume_type',
                  'memory', 'cpu', 'subnet', 'flavor', 'data_volumes')
        extra_kwargs = {
            'cluster': {'lookup_field': 'uuid', 'view_name': 'rancher-cluster-detail'}
        }

    def validate(self, attrs):
        attrs = super(CreateNodeSerializer, self).validate(attrs)
        cluster = attrs.get('cluster')
        spl = cluster.service_project_link
        node = attrs
        utils.expand_added_nodes([node], spl, cluster.tenant_settings, cluster.name)
        return attrs


class ClusterImportableSerializer(serializers.Serializer):
    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='rancher-spl-detail',
        queryset=models.RancherServiceProjectLink.objects.all(),
        write_only=True)

    name = serializers.CharField(read_only=True)
    backend_id = serializers.CharField(source="id", read_only=True)
    type = serializers.SerializerMethodField()
    extra = serializers.SerializerMethodField()

    def get_type(self, cluster):
        return 'Rancher.Cluster'

    def get_extra(self, cluster):
        spec = cluster.get('appliedSpec', {})
        config = spec.get('rancherKubernetesEngineConfig', {})
        backend_nodes = config.get('nodes', [])
        return [
            {
                'name': 'Number of nodes',
                'value': len(backend_nodes),
            },
            {
                'name': 'Created at',
                'value': cluster.get('created'),
            },
        ]


class ClusterImportSerializer(ClusterImportableSerializer):
    backend_id = serializers.CharField(write_only=True)

    @transaction.atomic
    def create(self, validated_data):
        service_project_link = validated_data['service_project_link']
        backend_id = validated_data['backend_id']

        if models.Cluster.objects.filter(
            service_project_link__service__settings=service_project_link.service.settings,
            backend_id=backend_id
        ).exists():
            raise serializers.ValidationError({'backend_id': _('Cluster has been imported already.')})

        try:
            backend = service_project_link.get_backend()
            cluster = backend.import_cluster(backend_id, service_project_link=service_project_link)
        except exceptions.RancherException:
            raise serializers.ValidationError({
                'backend_id': _("Can't import cluster with ID %s") % validated_data['backend_id']
            })

        return cluster


class LinkOpenstackSerializer(serializers.Serializer):
    instance = serializers.HyperlinkedRelatedField(
        view_name='openstacktenant-instance-detail',
        queryset=openstack_tenant_models.Instance.objects.all(),
        lookup_field='uuid',
        write_only=True)


class CatalogSerializer(serializers.HyperlinkedModelSerializer):
    scope = core_serializers.GenericRelatedField()

    class Meta:
        model = models.Catalog
        fields = ('uuid', 'url', 'created', 'modified', 'name', 'description',
                  'catalog_url', 'branch', 'commit', 'runtime_state', 'scope', 'scope_type')
        read_only_fields = ('runtime_state', 'commit')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid', 'view_name': 'rancher-catalog-detail'},
        }


class CatalogCreateSerializer(CatalogSerializer):
    class Meta(CatalogSerializer.Meta):
        fields = CatalogSerializer.Meta.fields + ('username', 'password')


class CatalogUpdateSerializer(CatalogCreateSerializer):
    class Meta(CatalogSerializer.Meta):
        read_only_fields = CatalogSerializer.Meta.read_only_fields + ('scope',)



from waldur_core.logging.loggers import EventLogger, event_logger
from waldur_core.structure import models as structure_models

from . import models


class ResourceActionEventLogger(EventLogger):
    resource = structure_models.NewResource
    action_details = dict

    class Meta:
        event_types = (
            'resource_pull_scheduled',
            'resource_pull_succeeded',
            'resource_pull_failed',
            # volume
            'resource_attach_scheduled',
            'resource_attach_succeeded',
            'resource_attach_failed',

            'resource_detach_scheduled',
            'resource_detach_succeeded',
            'resource_detach_failed',

            'resource_extend_scheduled',
            'resource_extend_succeeded',
            'resource_extend_failed',

            # instance
            'resource_update_security_groups_scheduled',
            'resource_update_security_groups_succeeded',
            'resource_update_security_groups_failed',

            'resource_change_flavor_scheduled',
            'resource_change_flavor_succeeded',
            'resource_change_flavor_failed',

            'resource_assign_floating_ip_scheduled',
            'resource_assign_floating_ip_succeeded',
            'resource_assign_floating_ip_failed',

            'resource_stop_scheduled',
            'resource_stop_succeeded',
            'resource_stop_failed',

            'resource_start_scheduled',
            'resource_start_succeeded',
            'resource_start_failed',

            'resource_restart_scheduled',
            'resource_restart_succeeded',
            'resource_restart_failed',

            'resource_extend_volume_scheduled',
            'resource_extend_volume_succeeded',
            'resource_extend_volume_failed',

            'resource_retype_scheduled',
            'resource_retype_succeeded',
            'resource_retype_failed',

            'resource_unassign_floating_ip_scheduled',
            'resource_unassign_floating_ip_succeeded',
            'resource_unassign_floating_ip_failed',

            'resource_update_internal_ips_scheduled',
            'resource_update_internal_ips_succeeded',
            'resource_update_internal_ips_failed',

            'resource_update_floating_ips_scheduled',
            'resource_update_floating_ips_succeeded',
            'resource_update_floating_ips_failed',
        )
        event_groups = {'resources': event_types}

    @staticmethod
    def get_scopes(event_context):
        resource = event_context['resource']
        project = resource.service_project_link.project
        return {resource, project, project.customer}


class BackupScheduleEventLogger(EventLogger):
    resource = models.Instance
    backup_schedule = models.BackupSchedule

    class Meta:
        event_types = (
            'resource_backup_schedule_created',
            'resource_backup_schedule_deleted',
            'resource_backup_schedule_activated',
            'resource_backup_schedule_deactivated',
            'resource_backup_schedule_cleaned_up',
        )
        event_groups = {'resources': event_types}

    @staticmethod
    def get_scopes(event_context):
        return ResourceActionEventLogger.get_scopes(event_context)


class SnapshotScheduleEventLogger(EventLogger):
    resource = models.Volume
    snapshot_schedule = models.SnapshotSchedule

    class Meta:
        event_types = (
            'resource_snapshot_schedule_created',
            'resource_snapshot_schedule_deleted',
            'resource_snapshot_schedule_activated',
            'resource_snapshot_schedule_deactivated',
            'resource_snapshot_schedule_cleaned_up',
        )
        event_groups = {'resources': event_types}

    @staticmethod
    def get_scopes(event_context):
        return ResourceActionEventLogger.get_scopes(event_context)


class BackupEventLogger(EventLogger):
    resource = models.Instance

    class Meta:
        event_types = ('resource_backup_creation_scheduled',
                       'resource_backup_creation_succeeded',
                       'resource_backup_creation_failed',
                       'resource_backup_restoration_scheduled',
                       'resource_backup_restoration_succeeded',
                       'resource_backup_restoration_failed',
                       'resource_backup_deletion_scheduled',
                       'resource_backup_deletion_succeeded',
                       'resource_backup_deletion_failed',
                       'resource_backup_schedule_creation_succeeded',
                       'resource_backup_schedule_update_succeeded',
                       'resource_backup_schedule_deletion_succeeded',
                       'resource_backup_schedule_activated',
                       'resource_backup_schedule_deactivated')

    @staticmethod
    def get_scopes(event_context):
        return ResourceActionEventLogger.get_scopes(event_context)


event_logger.register('openstack_resource_action', ResourceActionEventLogger)
event_logger.register('openstack_backup_schedule', BackupScheduleEventLogger)
event_logger.register('openstack_snapshot_schedule', SnapshotScheduleEventLogger)
event_logger.register('openstack_backup', BackupEventLogger)



from rest_framework import serializers

from waldur_core.core import signals as core_signals
from waldur_openstack.openstack_tenant import serializers as openstack_serializers
from waldur_zabbix import models as zabbix_models
from waldur_zabbix import serializers as zabbix_serializers


class LinkSerializer(zabbix_serializers.ServiceProjectLinkSerializer):
    internal_ip = serializers.SerializerMethodField()
    service_settings = serializers.HyperlinkedRelatedField(
        source='service.settings',
        view_name='servicesettings-detail',
        read_only=True,
        lookup_field='uuid')
    service_settings_uuid = serializers.ReadOnlyField(source='service.settings.uuid')

    def get_internal_ip(self, link):
        scope = link.service.settings.scope
        if not scope:
            return
        if not hasattr(scope, 'internal_ips'):
            return
        # Note that we do not support multiple IPs per VM yet
        if isinstance(scope.internal_ips, list):
            return scope.internal_ips[0]
        return scope.internal_ips

    class Meta(zabbix_serializers.ServiceProjectLinkSerializer.Meta):
        fields = zabbix_serializers.ServiceProjectLinkSerializer.Meta.fields + (
            'internal_ip', 'service_settings', 'service_settings_uuid',
        )


class NestedHostSerializer(serializers.HyperlinkedModelSerializer):
    state = serializers.ReadOnlyField(source='get_state_display')

    class Meta(serializers.HyperlinkedModelSerializer):
        model = zabbix_models.Host
        fields = ('url', 'uuid', 'state')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid', 'view_name': 'zabbix-host-detail'},
        }


def get_zabbix_host(serializer, scope):
    host = zabbix_models.Host.objects.filter(scope=scope).last()
    if host:
        serializer = NestedHostSerializer(instance=host, context=serializer.context)
        return serializer.data


def add_zabbix_host(sender, fields, **kwargs):
    fields['zabbix_host'] = serializers.SerializerMethodField()
    setattr(sender, 'get_zabbix_host', get_zabbix_host)


core_signals.pre_serializer_fields.connect(
    sender=openstack_serializers.InstanceSerializer,
    receiver=add_zabbix_host,
)



from waldur_core.core import WaldurExtension


class MarketplaceSlurmExtension(WaldurExtension):

    class Settings:
        WALDUR_MARKETPLACE_SLURM = {
            'CATEGORY_UUID': None,
        }

    @staticmethod
    def django_app():
        return 'waldur_mastermind.marketplace_slurm'

    @staticmethod
    def is_assembly():
        return True



import collections
import os
import re

from jira import JIRA, JIRAError, utils
from jira.resources import Attachment
from requests import Request

PADDING = 3
CHARS_LIMIT = 255


def _get_filename(path):
    limit = CHARS_LIMIT - PADDING
    fname = os.path.basename(path)
    filename = fname.split('.')[0]
    filename_extension = fname.split('.')[1:]
    count = len('.'.join(filename_extension).encode('utf-8')) + 1 if filename_extension else 0
    char_limit = 0

    for char in filename:
        count += len(char.encode('utf-8'))
        if count > limit:
            break
        else:
            char_limit += 1

    if not char_limit:
        raise JIRAError('Attachment filename is very long.')

    tmp = [filename[:char_limit]]
    tmp.extend(filename_extension)
    filename = '.'.join(tmp)
    return filename


def _upload_file(manager, issue, upload_file, filename):
    # This method will fix original method jira.JIRA.add_attachment (jira/client.py line 591)
    url = manager._get_url('issue/' + str(issue) + '/attachments')
    files = {'file': (filename, upload_file), }
    headers = {'X-Atlassian-Token': 'nocheck', }
    req = Request('POST', url, headers=headers, files=files, auth=manager._session.auth)
    prepped = req.prepare()
    prepped.body = re.sub(b'filename\*=.*', b'filename="%s"\r' % filename.encode('utf-8'), prepped.body)
    r = manager._session.send(prepped)

    js = utils.json_loads(r)

    if not js or not isinstance(js, collections.Iterable):
        raise JIRAError("Unable to parse JSON: %s" % js)

    attachment = Attachment(manager._options, manager._session, js[0])

    if attachment.size == 0:
        raise JIRAError("Added empty attachment?!: r: %s\nattachment: %s" % (r, attachment))

    return attachment


def add_attachment(manager, issue, path):
    """
    Replace jira's method 'add_attachment' while don't well fixed this issue
    https://github.com/shazow/urllib3/issues/303
    And we need to set filename limit equaled 252 chars.
    :param manager: [jira.JIRA instance]
    :param issue: [jira.JIRA.resources.Issue instance]
    :param path: [string]
    :return: [jira.JIRA.resources.Attachment instance]
    """
    filename = _get_filename(path)

    with open(path, 'rb') as f:
        return _upload_file(manager, issue, f, filename)


JIRA.waldur_add_attachment = add_attachment



from . import views


def register_in(router):
    router.register(r'azure', views.AzureServiceViewSet, basename='azure')
    router.register(r'azure-images', views.ImageViewSet, basename='azure-image')
    router.register(r'azure-sizes', views.SizeViewSet, basename='azure-size')
    router.register(r'azure-locations', views.LocationViewSet, basename='azure-location')
    router.register(r'azure-resource-groups', views.ResourceGroupViewSet, basename='azure-resource-group')
    router.register(r'azure-virtualmachines', views.VirtualMachineViewSet, basename='azure-virtualmachine')
    router.register(r'azure-public-ips', views.PublicIPViewSet, basename='azure-public-ip')
    router.register(r'azure-sql-servers', views.SQLServerViewSet, basename='azure-sql-server')
    router.register(r'azure-sql-databases', views.SQLDatabaseViewSet, basename='azure-sql-database')
    router.register(r'azure-service-project-link',
                    views.AzureServiceProjectLinkViewSet, basename='azure-spl')



from django.contrib import admin
from django.utils.translation import ugettext_lazy as _
from django.shortcuts import redirect
from django.urls import reverse

from waldur_core.core import admin as core_admin
from waldur_core.structure import admin as structure_admin

from . import models, tasks


class RancherUserClusterLinkInline(admin.TabularInline):
    model = models.RancherUserClusterLink


class RancherUserAdmin(core_admin.ExtraActionsMixin, core_admin.ReadOnlyAdminMixin, admin.ModelAdmin):
    list_display = ('__str__', 'settings', 'is_active')

    inlines = [
        RancherUserClusterLinkInline,
    ]

    def get_extra_actions(self):
        return [
            self.sync_users,
        ]

    def sync_users(self, request):
        tasks.sync_users.delay()
        self.message_user(request, _('Users\' synchronization has been scheduled.'))
        return redirect(reverse('admin:waldur_rancher_rancheruser_changelist'))


class CatalogAdmin(admin.ModelAdmin):
    list_display = ('name', 'scope_type', 'scope_name', 'catalog_url')

    def scope_name(self, obj):
        return obj.scope.name


admin.site.register(models.RancherService, structure_admin.ServiceAdmin)
admin.site.register(models.RancherServiceProjectLink, structure_admin.ServiceProjectLinkAdmin)
admin.site.register(models.Cluster)
admin.site.register(models.Node)
admin.site.register(models.RancherUser, RancherUserAdmin)
admin.site.register(models.Catalog, CatalogAdmin)



from io import BytesIO
import datetime
import os
import tarfile

from django.apps import apps
from django.core.files.base import ContentFile

from waldur_core.core.utils import datetime_to_timestamp
from waldur_core.logging.loggers import LoggableMixin


def get_loggable_models():
    return [model for model in apps.get_models() if issubclass(model, LoggableMixin)]


def get_scope_types_mapping():
    return {str(m._meta): m for m in get_loggable_models()}


def get_reverse_scope_types_mapping():
    return {m: str(m._meta) for m in get_loggable_models()}


def create_report_archive(log_directory, interval):
    """
    Create tar.gz archive from files in directory filtered by time delta.
    :param log_directory: directory with log files, for example, /var/log/waldur/
    :param interval: time delta, for example, datetime.timedelta(days=7)
    files older that specified interval are filtered out
    :return: ContentFile with gzipped archive
    """
    today = datetime.datetime.today()
    cutoff = datetime_to_timestamp(today - interval)

    log_filenames = []
    log_dir = log_directory
    for log_file in os.listdir(log_dir):
        full_path = os.path.join(log_dir, log_file)
        stat = os.stat(full_path)
        if stat.st_mtime > cutoff:
            log_filenames.append(full_path)

    stream = BytesIO()

    with tarfile.open(fileobj=stream, mode='w:gz') as archive:
        for filename in log_filenames:
            archive.add(filename)

    return ContentFile(stream.getvalue())



# -*- coding: utf-8 -*-

import django.utils.timezone
from django.db import migrations, models

import waldur_core.core.validators


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='ResourceItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name],
                                          verbose_name='name')),
                ('object_id', models.PositiveIntegerField()),
                ('value', models.FloatField()),
                ('content_type',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
            ],
        ),
        migrations.CreateModel(
            name='ResourceSla',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('object_id', models.PositiveIntegerField()),
                ('period', models.CharField(max_length=10)),
                ('value', models.DecimalField(blank=True, decimal_places=4, max_digits=11, null=True)),
                ('agreed_value', models.DecimalField(blank=True, decimal_places=4, max_digits=11, null=True)),
                ('content_type',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
            ],
        ),
        migrations.CreateModel(
            name='ResourceSlaStateTransition',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('object_id', models.PositiveIntegerField()),
                ('period', models.CharField(max_length=10)),
                ('timestamp', models.IntegerField()),
                ('state', models.BooleanField(default=False, help_text='If state is True resource became available')),
                ('content_type',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
            ],
        ),
        migrations.AlterUniqueTogether(
            name='resourceslastatetransition',
            unique_together=set([('timestamp', 'period', 'content_type', 'object_id')]),
        ),
        migrations.AlterUniqueTogether(
            name='resourcesla',
            unique_together=set([('period', 'content_type', 'object_id')]),
        ),
        migrations.AlterUniqueTogether(
            name='resourceitem',
            unique_together=set([('name', 'content_type', 'object_id')]),
        ),
    ]



from django import forms
from django.contrib import admin
from django.core.exceptions import ValidationError
from django.utils.translation import ungettext
from waldur_core.core.admin import JsonWidget

from waldur_core.core.admin import ExecutorAdminAction
from waldur_core.structure import admin as structure_admin

from . import executors, tasks
from .models import ZabbixServiceProjectLink, ZabbixService, Host, SlaHistory, SlaHistoryEvent, ITService


class SlaHistoryEventsInline(admin.TabularInline):
    model = SlaHistoryEvent
    fields = ('timestamp', 'state')
    ordering = ('timestamp', )
    extra = 1


class SlaHistoryAdmin(admin.ModelAdmin):
    inlines = (
        SlaHistoryEventsInline,
    )
    list_display = ('itservice', 'period', 'value')
    ordering = ('itservice', 'period')
    list_filter = ('period',)


class HostAdminForm(forms.ModelForm):
    class Meta:
        widgets = {
            'interface_parameters': JsonWidget(),
        }


class HostAdmin(structure_admin.ResourceAdmin):
    actions = ['pull_sla', 'pull']
    form = HostAdminForm

    # TODO: Rewrite with executor.
    def pull_sla(self, request, queryset):
        for host in queryset:
            tasks.pull_sla.delay(host.uuid.hex)

        tasks_scheduled = queryset.count()
        message = ungettext(
            'SLA pulling has been scheduled for one host',
            'SLA pulling has been scheduled for %(tasks_scheduled)d hosts',
            tasks_scheduled
        )
        message = message % {'tasks_scheduled': tasks_scheduled}

        self.message_user(request, message)

    pull_sla.short_description = "Pull SLAs for given Zabbix hosts"

    class Pull(ExecutorAdminAction):
        executor = executors.HostPullExecutor
        short_description = 'Pull'

        def validate(self, instance):
            if instance.state not in (Host.States.OK, Host.States.ERRED):
                raise ValidationError('Host has to be in stable state (OK or ERRED) to be pulled.')

    pull = Pull()


class ITServiceAdmin(structure_admin.ResourceAdmin):
    pass


admin.site.register(Host, HostAdmin)
admin.site.register(ITService, ITServiceAdmin)
admin.site.register(ZabbixService, structure_admin.ServiceAdmin)
admin.site.register(ZabbixServiceProjectLink, structure_admin.ServiceProjectLinkAdmin)
admin.site.register(SlaHistory, SlaHistoryAdmin)



from django.core.management.base import BaseCommand
from django.db import transaction

from waldur_mastermind.billing import models


class Command(BaseCommand):
    help = "Create or update price estimates based on invoices."

    def handle(self, *args, **options):
        with transaction.atomic():
            for model in models.PriceEstimate.get_estimated_models():
                for instance in model.objects.all():
                    estimate, _ = models.PriceEstimate.objects.get_or_create(scope=instance)
                    estimate.update_total()
                    estimate.save()



from waldur_mastermind.marketplace import models as marketplace_models
from waldur_mastermind.marketplace.utils import get_resource_state


def create_marketplace_resource_for_imported_cluster(sender, instance, offering=None, plan=None, **kwargs):
    if not offering or not plan:
        # When cluster is imported directly (ie without marketplace),
        # marketplace resources are not created.
        return
    resource = marketplace_models.Resource(
        project=instance.service_project_link.project,
        state=get_resource_state(instance.state),
        name=instance.name,
        scope=instance,
        created=instance.created,
        plan=plan,
        offering=offering,
    )

    resource.init_cost()
    resource.save()
    resource.init_quotas()



import logging


from waldur_core.core import models as core_models, tasks as core_tasks, utils as core_utils
from waldur_core.structure import (filters as structure_filters, permissions as structure_permissions,
                                   models as structure_models)
from waldur_openstack.openstack import apps

from .log import event_logger
from .models import Tenant


logger = logging.getLogger(__name__)


def remove_ssh_key_from_tenants(sender, structure, user, role, **kwargs):
    """ Delete user ssh keys from tenants that he does not have access now. """
    tenants = Tenant.objects.filter(**{sender.__name__.lower(): structure})
    ssh_keys = core_models.SshPublicKey.objects.filter(user=user)
    for tenant in tenants:
        if structure_permissions._has_admin_access(user, tenant.project):
            continue  # no need to delete ssh keys if user still have permissions for tenant.
        serialized_tenant = core_utils.serialize_instance(tenant)
        for key in ssh_keys:
            core_tasks.BackendMethodTask().delay(
                serialized_tenant, 'remove_ssh_key_from_tenant', key.name, key.fingerprint)


def remove_ssh_key_from_all_tenants_on_it_deletion(sender, instance, **kwargs):
    """ Delete key from all tenants that are accessible for user on key deletion. """
    ssh_key = instance
    user = ssh_key.user
    tenants = structure_filters.filter_queryset_for_user(Tenant.objects.all(), user)
    for tenant in tenants:
        if not structure_permissions._has_admin_access(user, tenant.project):
            continue
        serialized_tenant = core_utils.serialize_instance(tenant)
        core_tasks.BackendMethodTask().delay(
            serialized_tenant, 'remove_ssh_key_from_tenant', ssh_key.name, ssh_key.fingerprint)


def log_tenant_quota_update(sender, instance, created=False, **kwargs):
    quota = instance
    if created or not isinstance(quota.scope, Tenant):
        return

    if not quota.tracker.has_changed('limit'):
        return

    tenant = quota.scope
    new_value_representation = quota.scope.format_quota(quota.name, quota.limit)
    old_value_representation = quota.scope.format_quota(quota.name, quota.tracker.previous('limit'))
    event_logger.openstack_tenant_quota.info(
        '{quota_name} quota limit has been changed from %s to %s for tenant {tenant_name}.' %
        (old_value_representation, new_value_representation),
        event_type='openstack_tenant_quota_limit_updated',
        event_context={
            'quota': quota,
            'tenant': tenant,
            'limit': float(quota.limit),
            'old_limit': float(quota.tracker.previous('limit')),
        })


def update_service_settings_name(sender, instance, created=False, **kwargs):
    tenant = instance

    if created or not tenant.tracker.has_changed('name'):
        return

    try:
        service_settings = structure_models.ServiceSettings.objects.get(scope=tenant,
                                                                        type=apps.OpenStackConfig.service_name)
    except structure_models.ServiceSettings.DoesNotExist:
        return
    else:
        service_settings.name = tenant.name
        service_settings.save()



# Generated by Django 2.2.9 on 2020-02-12 15:19

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields
import waldur_core.core.fields
import waldur_core.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('waldur_rancher', '0010_rancher_user'),
    ]

    operations = [
        migrations.CreateModel(
            name='Catalog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('runtime_state', models.CharField(blank=True, max_length=150, verbose_name='runtime state')),
                ('object_id', models.PositiveIntegerField(null=True)),
                ('backend_id', models.CharField(blank=True, max_length=255)),
                ('catalog_url', models.URLField()),
                ('branch', models.CharField(max_length=255)),
                ('commit', models.CharField(blank=True, max_length=40)),
                ('username', models.CharField(blank=True, max_length=255)),
                ('password', models.CharField(blank=True, max_length=255)),
                ('content_type', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='contenttypes.ContentType')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



import functools
import logging

import digitalocean

from django.db import IntegrityError, transaction
from django.utils import dateparse

from waldur_core.core.models import SshPublicKey
from waldur_core.structure import ServiceBackend, ServiceBackendError, SupportedServices

from . import models


logger = logging.getLogger(__name__)


class DigitalOceanBackendError(ServiceBackendError):
    pass


class TokenScopeError(DigitalOceanBackendError):
    pass


class NotFoundError(DigitalOceanBackendError):
    pass


class UnauthorizedError(DigitalOceanBackendError):
    pass


def digitalocean_error_handler(func):
    """
    Convert DigitalOcean exception to specific classes based on text message.
    It should be applied to functions which directly call `manager`.
    """
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        error_messages = {
            'You do not have access for the attempted action.': TokenScopeError,
            'The resource you were accessing could not be found.': NotFoundError,
            'Unable to authenticate you.': UnauthorizedError
        }
        logger.debug('About to execute DO backend method `%s`' % func.__name__)
        try:
            return func(*args, **kwargs)
        except digitalocean.DataReadError as e:
            raise error_messages.get(str(e))
    return wrapped


class DigitalOceanBackend(ServiceBackend):
    """ Waldur interface to Digital Ocean API.
        https://developers.digitalocean.com/documentation/v2/
        https://github.com/koalalorenzo/python-digitalocean
    """

    def __init__(self, settings):
        self.settings = settings
        self.manager = digitalocean.Manager(token=settings.token)

    def pull_resources(self):
        self.pull_droplets()

    @digitalocean_error_handler
    def create_droplet(self, droplet, backend_region_id=None, backend_image_id=None,
                       backend_size_id=None, ssh_key_uuid=None):

        if ssh_key_uuid:
            ssh_key = SshPublicKey.objects.get(uuid=ssh_key_uuid)
            backend_ssh_key = self.get_or_create_ssh_key(ssh_key)

            droplet.key_name = ssh_key.name
            droplet.key_fingerprint = ssh_key.fingerprint

        backend_droplet = digitalocean.Droplet(
            token=self.manager.token,
            name=droplet.name,
            user_data=droplet.user_data,
            region=backend_region_id,
            image=backend_image_id,
            size_slug=backend_size_id,
            ssh_keys=[backend_ssh_key.id] if ssh_key_uuid else [])
        backend_droplet.create()

        action_id = backend_droplet.action_ids[-1]
        droplet.backend_id = backend_droplet.id
        droplet.save()
        return action_id

    @digitalocean_error_handler
    def destroy(self, droplet):
        backend_droplet = self.get_droplet(droplet.backend_id)
        backend_droplet.destroy()
        droplet.decrease_backend_quotas_usage()

    @digitalocean_error_handler
    def start(self, droplet):
        backend_droplet = self.get_droplet(droplet.backend_id)
        action = backend_droplet.power_on()
        return action['action']['id']

    @digitalocean_error_handler
    def stop(self, droplet):
        backend_droplet = self.get_droplet(droplet.backend_id)
        action = backend_droplet.shutdown()
        return action['action']['id']

    @digitalocean_error_handler
    def restart(self, droplet):
        backend_droplet = self.get_droplet(droplet.backend_id)
        action = backend_droplet.reboot()
        return action['action']['id']

    @digitalocean_error_handler
    def resize(self, droplet, backend_size_id=None, disk=None):
        backend_droplet = self.get_droplet(droplet.backend_id)
        action = backend_droplet.resize(new_size_slug=backend_size_id, disk=disk)
        return action['action']['id']

    @digitalocean_error_handler
    def remove_ssh_key(self, name, fingerprint):
        try:
            backend_ssh_key = self.pull_ssh_key(name, fingerprint)
        except NotFoundError:
            pass  # no need to perform any action if key doesn't exist at backend
        else:
            backend_ssh_key.destroy()

    def ping(self, raise_exception=False):
        tries_count = 3
        for _ in range(tries_count):
            try:
                self.manager.get_account()
            except digitalocean.DataReadError as e:
                if raise_exception:
                    raise DigitalOceanBackendError(e)
            else:
                return True
        return False

    def ping_resource(self, droplet):
        try:
            self.get_droplet(droplet.backend_id)
        except NotFoundError:
            return False
        else:
            return True

    def pull_service_properties(self):
        self.pull_regions()
        self.pull_images()
        self.pull_sizes()

    def has_global_properties(self):
        properties = (models.Region, models.Image, models.Size)
        return all(model.objects.count() > 0 for model in properties)

    @transaction.atomic
    def pull_regions(self):
        cur_regions = self._get_current_properties(models.Region)
        for backend_region in self.get_all_regions():
            if backend_region.available:
                cur_regions.pop(backend_region.slug, None)
                try:
                    models.Region.objects.update_or_create(
                        backend_id=backend_region.slug,
                        defaults={'name': backend_region.name})
                except IntegrityError:
                    logger.warning(
                        'Could not create DigitalOcean region with id %s due to concurrent update',
                        backend_region.slug)

        models.Region.objects.filter(backend_id__in=cur_regions.keys()).delete()

    @transaction.atomic
    def pull_images(self):
        cur_images = self._get_current_properties(models.Image)
        for backend_image in self.get_all_images():
            cur_images.pop(str(backend_image.id), None)
            try:
                image, _ = models.Image.objects.update_or_create(
                    backend_id=backend_image.id,
                    defaults={
                        'name': '{} {}'.format(backend_image.distribution, backend_image.name),
                        'type': backend_image.type,
                        'distribution': backend_image.distribution,
                        'is_official': backend_image.slug is not None,
                        'min_disk_size': self.gb2mb(backend_image.min_disk_size),
                        'created_at': dateparse.parse_datetime(backend_image.created_at)
                    })
                self._update_entity_regions(image, backend_image)
            except IntegrityError:
                logger.warning(
                    'Could not create DigitalOcean image with id %s due to concurrent update',
                    backend_image.id)

        models.Image.objects.filter(backend_id__in=cur_images.keys()).delete()

    @transaction.atomic
    def pull_sizes(self):
        cur_sizes = self._get_current_properties(models.Size)
        for backend_size in self.get_all_sizes():
            cur_sizes.pop(backend_size.slug, None)
            try:
                size, _ = models.Size.objects.update_or_create(
                    backend_id=backend_size.slug,
                    defaults={
                        'name': backend_size.slug,
                        'cores': backend_size.vcpus,
                        'ram': backend_size.memory,
                        'disk': self.gb2mb(backend_size.disk),
                        'transfer': int(self.tb2mb(backend_size.transfer)),
                        'price': backend_size.price_hourly})
                self._update_entity_regions(size, backend_size)
            except IntegrityError:
                logger.warning(
                    'Could not create DigitalOcean size with id %s due to concurrent update',
                    backend_size.slug)

        models.Size.objects.filter(backend_id__in=cur_sizes.keys()).delete()

    @transaction.atomic
    def pull_droplets(self):
        backend_droplets = {str(droplet.id): droplet for droplet in self.get_all_droplets()}
        backend_ids = set(backend_droplets.keys())

        nc_droplets = models.Droplet.objects.filter(service_project_link__service__settings=self.settings)
        nc_droplets = {droplet.backend_id: droplet for droplet in nc_droplets}
        nc_ids = set(nc_droplets.keys())

        # Mark stale droplets as erred if they are removed from the backend
        for droplet_id in nc_ids - backend_ids:
            nc_droplet = nc_droplets[droplet_id]
            nc_droplet.set_erred()
            nc_droplet.save()

        # Update state of matching droplets
        for droplet_id in nc_ids & backend_ids:
            backend_droplet = backend_droplets[droplet_id]
            nc_droplet = nc_droplets[droplet_id]
            nc_droplet.state, nc_droplet.runtime_state = self._get_droplet_states(backend_droplet)
            nc_droplet.image_name = self.format_image_name(backend_droplet.image)
            nc_droplet.save()

    def _get_droplet_states(self, droplet):
        States = models.Droplet.States
        RuntimeStates = models.Droplet.RuntimeStates

        digitalocean_to_waldur = {
            'new': (States.CREATING, 'provisioning'),
            'active': (States.OK, RuntimeStates.ONLINE),
            'off': (States.OK, RuntimeStates.OFFLINE),
            'archive': (States.OK, 'archive'),
        }

        return digitalocean_to_waldur.get(droplet.status, (States.ERRED, 'error'))

    def format_image_name(self, backend_image):
        return '{} {}'.format(backend_image['distribution'], backend_image['name'])

    @digitalocean_error_handler
    def get_droplet(self, backend_droplet_id):
        return self.manager.get_droplet(backend_droplet_id)

    def get_monthly_cost_estimate(self, droplet):
        backend_droplet = self.get_droplet(droplet.backend_id)
        return backend_droplet.size['price_monthly']

    def get_resources_for_import(self):
        cur_droplets = models.Droplet.objects.all().values_list('backend_id', flat=True)
        statuses = ('active', 'off')
        droplets = self.get_all_droplets()
        return [{
            'id': droplet.id,
            'name': droplet.name,
            'created_at': droplet.created_at,
            'cores': droplet.vcpus,
            'ram': droplet.memory,
            'disk': self.gb2mb(droplet.disk),
            'flavor_name': droplet.size_slug,
            'resource_type': SupportedServices.get_name_for_model(models.Droplet)
        } for droplet in droplets
            if str(droplet.id) not in cur_droplets and droplet.status in statuses]

    def import_droplet(self, backend_droplet_id, service_project_link=None, save=True):
        backend_droplet = self.get_droplet(backend_droplet_id)
        droplet = models.Droplet()
        droplet.backend_id = backend_droplet_id
        droplet.name = backend_droplet.name
        droplet.image_name = self.format_image_name(backend_droplet.image)
        droplet.size_name = backend_droplet.size_slug
        droplet.cores = backend_droplet.vcpus
        droplet.ram = backend_droplet.memory
        droplet.disk = self.gb2mb(backend_droplet.disk)
        droplet.transfer = self.tb2mb(backend_droplet.size['transfer'])
        droplet.ip_address = backend_droplet.ip_address
        droplet.created = dateparse.parse_datetime(backend_droplet.created_at)
        state, runtime_state = self._get_droplet_states(backend_droplet)
        droplet.runtime_state = runtime_state
        if service_project_link and save:
            droplet.state = state
            droplet.service_project_link = service_project_link
            droplet.save()
        return droplet

    def get_managed_resources(self):
        try:
            ids = [droplet.id for droplet in self.get_all_droplets()]
            return models.Droplet.objects.filter(backend_id__in=ids)
        except DigitalOceanBackendError:
            return []

    @digitalocean_error_handler
    def get_all_droplets(self):
        return self.manager.get_all_droplets()

    @digitalocean_error_handler
    def get_all_regions(self):
        return self.manager.get_all_regions()

    @digitalocean_error_handler
    def get_all_images(self):
        return self.manager.get_all_images()

    @digitalocean_error_handler
    def get_all_sizes(self):
        return [size for size in self.manager.get_all_sizes() if size.slug]

    def get_or_create_ssh_key(self, ssh_key):
        try:
            backend_ssh_key = self.pull_ssh_key(ssh_key.name, ssh_key.fingerprint)
        except NotFoundError:
            backend_ssh_key = self.push_ssh_key(ssh_key)
        return backend_ssh_key

    @digitalocean_error_handler
    def push_ssh_key(self, ssh_key):
        backend_ssh_key = digitalocean.SSHKey(
            token=self.manager.token,
            name=ssh_key.name,
            public_key=ssh_key.public_key)

        backend_ssh_key.create()
        return backend_ssh_key

    @digitalocean_error_handler
    def pull_ssh_key(self, name, fingerprint):
        backend_ssh_key = digitalocean.SSHKey(
            token=self.manager.token,
            fingerprint=fingerprint,
            name=name,
            id=None)

        backend_ssh_key.load()
        return backend_ssh_key

    def _get_current_properties(self, model):
        return {p.backend_id: p for p in model.objects.all()}

    def _update_entity_regions(self, entity, backend_entity):
        all_regions = set(entity.regions.all())
        actual_regions = set(models.Region.objects.filter(backend_id__in=backend_entity.regions))

        entity.regions.add(*(actual_regions - all_regions))
        entity.regions.remove(*(all_regions - actual_regions))



# -*- coding: utf-8 -*-
# Generated by Django 1.11.3 on 2017-09-06 07:39
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import django_fsm
import model_utils.fields
import waldur_core.core.fields
import waldur_core.core.models
import waldur_core.logging.loggers


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('structure', '0001_squashed_0054'),
    ]

    operations = [
        migrations.CreateModel(
            name='Invoice',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('state', models.CharField(choices=[('DRAFT', 'Draft'), ('SENT', 'Sent'), ('PAID', 'Paid'), ('MARKED_AS_PAID', 'Marked as paid'), ('CANCELLED', 'Cancelled'), ('REFUNDED', 'Refunded'), ('PARTIALLY_REFUNDED', 'Partially refunded'), ('MARKED_AS_REFUNDED', 'Marked as refunded'), ('UNPAID', 'Unpaid'), ('PAYMENT_PENDING', 'Payment pending')], default='DRAFT', max_length=30)),
                ('invoice_date', models.DateField()),
                ('end_date', models.DateField()),
                ('pdf', models.FileField(blank=True, null=True, upload_to='paypal-invoices')),
                ('number', models.CharField(max_length=30)),
                ('tax_percent', models.DecimalField(decimal_places=2, default=0, max_digits=4, validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(100)])),
                ('backend_id', models.CharField(blank=True, max_length=128)),
                ('issuer_details', waldur_core.core.fields.JSONField(blank=True, default={}, help_text='Stores data about invoice issuer')),
                ('payment_details', waldur_core.core.fields.JSONField(blank=True, default={}, help_text='Stores data about customer payment details')),
                ('month', models.PositiveSmallIntegerField(validators=[django.core.validators.MinValueValidator(1), django.core.validators.MaxValueValidator(12)])),
                ('year', models.PositiveSmallIntegerField()),
                ('customer', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='paypal_invoices', to='structure.Customer')),
            ],
            options={
                'ordering': ['-invoice_date'],
            },
            bases=(waldur_core.logging.loggers.LoggableMixin, models.Model, waldur_core.core.models.BackendModelMixin),
        ),
        migrations.CreateModel(
            name='InvoiceItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('price', models.DecimalField(decimal_places=2, max_digits=9)),
                ('tax', models.DecimalField(decimal_places=2, default=0, max_digits=9)),
                ('unit_price', models.DecimalField(decimal_places=2, max_digits=9)),
                ('quantity', models.PositiveIntegerField(default=0)),
                ('unit_of_measure', models.CharField(choices=[('QUANTITY', 'Quantity'), ('HOURS', 'Hours'), ('AMOUNT', 'Amount')], default='HOURS', max_length=30)),
                ('name', models.CharField(max_length=255)),
                ('start', models.DateTimeField(null=True)),
                ('end', models.DateTimeField(null=True)),
                ('invoice', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='items', to='waldur_paypal.Invoice')),
            ],
            options={
                'ordering': ['invoice', '-start'],
            },
        ),
        migrations.CreateModel(
            name='Payment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('state', django_fsm.FSMIntegerField(choices=[(0, 'Initial'), (1, 'Created'), (2, 'Approved'), (4, 'Erred')], default=0)),
                ('amount', models.DecimalField(decimal_places=2, max_digits=9)),
                ('tax', models.DecimalField(decimal_places=2, default=0, max_digits=9)),
                ('backend_id', models.CharField(max_length=255, null=True)),
                ('token', models.CharField(max_length=255, null=True)),
                ('approval_url', models.URLField()),
                ('customer', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structure.Customer')),
            ],
            options={
                'ordering': ['-modified'],
            },
            bases=(waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
    ]



# Generated by Django 2.2.7 on 2019-12-02 09:26

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0006_orderitem_output'),
    ]

    operations = [
        migrations.AddField(
            model_name='offering',
            name='enable_dynamic_components',
            field=models.BooleanField(default=False, help_text='For example, allow to use offering components related to volume types.'),
        ),
    ]



from django.db import models
from django.utils.translation import ugettext_lazy as _
from model_utils import FieldTracker

from waldur_core.core import models as core_models
from waldur_core.structure import models as structure_models


class VMwareService(structure_models.Service):
    projects = models.ManyToManyField(
        structure_models.Project,
        related_name='+',
        through='VMwareServiceProjectLink'
    )

    class Meta:
        unique_together = ('customer', 'settings')
        verbose_name = _('VMware provider')
        verbose_name_plural = _('VMware providers')

    @classmethod
    def get_url_name(cls):
        return 'vmware'


class VMwareServiceProjectLink(structure_models.ServiceProjectLink):

    service = models.ForeignKey(on_delete=models.CASCADE, to=VMwareService)

    class Meta(structure_models.ServiceProjectLink.Meta):
        verbose_name = _('VMware provider project link')
        verbose_name_plural = _('VMware provider project links')

    @classmethod
    def get_url_name(cls):
        return 'vmware-spl'


class VirtualMachineMixin(models.Model):
    class Meta:
        abstract = True

    guest_os = models.CharField(max_length=50, help_text=_('Defines the valid guest operating system '
                                                           'types used for configuring a virtual machine'))
    cores = models.PositiveSmallIntegerField(default=0, help_text=_('Number of cores in a VM'))
    cores_per_socket = models.PositiveSmallIntegerField(default=1, help_text=_('Number of cores per socket in a VM'))
    ram = models.PositiveIntegerField(default=0, help_text=_('Memory size in MiB'), verbose_name=_('RAM'))
    disk = models.PositiveIntegerField(default=0, help_text=_('Disk size in MiB'))


class VirtualMachine(VirtualMachineMixin,
                     core_models.RuntimeStateMixin,
                     structure_models.NewResource):
    service_project_link = models.ForeignKey(
        VMwareServiceProjectLink,
        related_name='+',
        on_delete=models.PROTECT
    )

    class RuntimeStates:
        POWERED_OFF = 'POWERED_OFF'
        POWERED_ON = 'POWERED_ON'
        SUSPENDED = 'SUSPENDED'

        CHOICES = (
            (POWERED_OFF, 'Powered off'),
            (POWERED_ON, 'Powered on'),
            (SUSPENDED, 'Suspended'),
        )

    class GuestPowerStates:
        RUNNING = 'RUNNING'
        SHUTTING_DOWN = 'SHUTTING_DOWN'
        RESETTING = 'RESETTING'
        STANDBY = 'STANDBY'
        NOT_RUNNING = 'NOT_RUNNING'
        UNAVAILABLE = 'UNAVAILABLE'

        CHOICES = (
            (RUNNING, 'Running'),
            (SHUTTING_DOWN, 'Shutting down'),
            (RESETTING, 'Resetting'),
            (STANDBY, 'Standby'),
            (NOT_RUNNING, 'Not running'),
            (UNAVAILABLE, 'Unavailable'),
        )

    class ToolsStates:
        STARTING = 'STARTING'
        RUNNING = 'RUNNING'
        NOT_RUNNING = 'NOT_RUNNING'

        CHOICES = (
            (STARTING, 'Starting'),
            (RUNNING, 'Running'),
            (NOT_RUNNING, 'Not running'),
        )

    template = models.ForeignKey('Template', null=True, on_delete=models.SET_NULL)
    cluster = models.ForeignKey('Cluster', null=True, on_delete=models.SET_NULL)
    datastore = models.ForeignKey('Datastore', null=True, on_delete=models.SET_NULL)
    folder = models.ForeignKey('Folder', null=True, on_delete=models.SET_NULL)
    networks = models.ManyToManyField('Network', blank=True)
    guest_power_enabled = models.BooleanField(
        default=False,
        help_text='Flag indicating if the virtual machine is ready to process soft power operations.'
    )
    guest_power_state = models.CharField(
        'The power state of the guest operating system.',
        max_length=150,
        blank=True,
        choices=GuestPowerStates.CHOICES,
    )
    tools_installed = models.BooleanField(default=False)
    tools_state = models.CharField(
        'Current running status of VMware Tools running in the guest operating system.',
        max_length=50,
        blank=True,
        choices=ToolsStates.CHOICES,
    )
    tracker = FieldTracker()

    @classmethod
    def get_backend_fields(cls):
        return super(VirtualMachine, cls).get_backend_fields() + (
            'runtime_state', 'cores', 'cores_per_socket', 'ram', 'disk',
            'tools_installed', 'tools_state',
        )

    @classmethod
    def get_url_name(cls):
        return 'vmware-virtual-machine'

    @property
    def total_disk(self):
        return self.disks.aggregate(models.Sum('size'))['size__sum'] or 0

    def __str__(self):
        return self.name


class Port(core_models.RuntimeStateMixin, structure_models.NewResource):
    service_project_link = models.ForeignKey(
        VMwareServiceProjectLink,
        related_name='+',
        on_delete=models.PROTECT
    )
    vm = models.ForeignKey(on_delete=models.CASCADE, to=VirtualMachine)
    network = models.ForeignKey(on_delete=models.CASCADE, to='Network')
    mac_address = models.CharField(max_length=32, blank=True, verbose_name=_('MAC address'))

    @classmethod
    def get_backend_fields(cls):
        return super(Port, cls).get_backend_fields() + ('name', 'mac_address')

    @classmethod
    def get_url_name(cls):
        return 'vmware-port'

    def __str__(self):
        return self.name


class Disk(structure_models.NewResource):
    service_project_link = models.ForeignKey(
        VMwareServiceProjectLink,
        related_name='+',
        on_delete=models.PROTECT
    )

    size = models.PositiveIntegerField(help_text=_('Size in MiB'))
    vm = models.ForeignKey(on_delete=models.CASCADE, to=VirtualMachine, related_name='disks')

    @classmethod
    def get_url_name(cls):
        return 'vmware-disk'

    def __str__(self):
        return self.name

    @classmethod
    def get_backend_fields(cls):
        return super(Disk, cls).get_backend_fields() + ('name', 'size')


class Template(VirtualMachineMixin,
               core_models.DescribableMixin,
               structure_models.ServiceProperty):
    created = models.DateTimeField()
    modified = models.DateTimeField()

    @classmethod
    def get_url_name(cls):
        return 'vmware-template'

    def __str__(self):
        return self.name


class Cluster(structure_models.ServiceProperty):
    @classmethod
    def get_url_name(cls):
        return 'vmware-cluster'

    def __str__(self):
        return '%s / %s' % (self.settings, self.name)


class CustomerCluster(models.Model):
    customer = models.ForeignKey(structure_models.Customer, on_delete=models.CASCADE)
    cluster = models.ForeignKey('Cluster', on_delete=models.CASCADE)

    def __str__(self):
        return '%s / %s' % (self.customer, self.cluster)

    class Meta:
        unique_together = ('customer', 'cluster')


class Network(structure_models.ServiceProperty):
    type = models.CharField(max_length=255)

    @classmethod
    def get_url_name(cls):
        return 'vmware-network'

    def __str__(self):
        return '%s / %s' % (self.settings, self.name)


class CustomerNetwork(models.Model):
    # This model allows to specify allowed networks for VM provision
    customer = models.ForeignKey(structure_models.Customer, on_delete=models.CASCADE)
    network = models.ForeignKey('Network', on_delete=models.CASCADE)

    def __str__(self):
        return '%s / %s' % (self.customer, self.network)

    class Meta:
        unique_together = ('customer', 'network')


class CustomerNetworkPair(models.Model):
    # This model allows to specify allowed networks for existing VM NIC provision
    customer = models.ForeignKey(structure_models.Customer, on_delete=models.CASCADE)
    network = models.ForeignKey('Network', on_delete=models.CASCADE)

    def __str__(self):
        return '%s / %s' % (self.customer, self.network)

    class Meta:
        unique_together = ('customer', 'network')


class Datastore(structure_models.ServiceProperty):
    type = models.CharField(max_length=255)
    capacity = models.PositiveIntegerField(help_text="Capacity, in MB.", null=True, blank=True)
    free_space = models.PositiveIntegerField(help_text="Available space, in MB.", null=True, blank=True)

    @classmethod
    def get_url_name(cls):
        return 'vmware-datastore'

    def __str__(self):
        return '%s / %s' % (self.settings, self.name)


class CustomerDatastore(models.Model):
    customer = models.ForeignKey(structure_models.Customer, on_delete=models.CASCADE)
    datastore = models.ForeignKey('Datastore', on_delete=models.CASCADE)

    def __str__(self):
        return '%s / %s' % (self.customer, self.datastore)

    class Meta:
        unique_together = ('customer', 'datastore')


class Folder(structure_models.ServiceProperty):

    def __str__(self):
        return '%s / %s' % (self.settings, self.name)

    @classmethod
    def get_url_name(cls):
        return 'vmware-folder'


class CustomerFolder(models.Model):
    customer = models.ForeignKey(structure_models.Customer, on_delete=models.CASCADE)
    folder = models.ForeignKey('Folder', on_delete=models.CASCADE)

    def __str__(self):
        return '%s / %s' % (self.customer, self.folder)

    class Meta:
        unique_together = ('customer', 'folder')



import datetime
from django.conf import settings
from rest_framework import serializers
from rest_framework.reverse import reverse

from waldur_core.core import serializers as core_serializers
from waldur_core.core import utils as core_utils
from waldur_mastermind.common.utils import quantize_price

from . import models, utils


class InvoiceItemSerializer(serializers.HyperlinkedModelSerializer):
    tax = serializers.DecimalField(max_digits=15, decimal_places=7)
    total = serializers.DecimalField(max_digits=15, decimal_places=7)
    factor = serializers.ReadOnlyField(source='get_factor')

    scope_type = serializers.SerializerMethodField()
    scope_uuid = serializers.SerializerMethodField()

    class Meta:
        model = models.InvoiceItem
        fields = ('name', 'price', 'tax', 'total', 'unit_price', 'unit', 'factor',
                  'start', 'end', 'product_code', 'article_code', 'project_name', 'project_uuid',
                  'scope_type', 'scope_uuid',)

    def get_scope_type(self, item):
        # It should be implemented by inherited class
        return

    def get_scope_uuid(self, item):
        # It should be implemented by inherited class
        return


class GenericItemSerializer(InvoiceItemSerializer):
    details = serializers.JSONField()

    class Meta(InvoiceItemSerializer.Meta):
        model = models.InvoiceItem
        fields = InvoiceItemSerializer.Meta.fields + ('quantity', 'details', 'usage_days',)

    def get_scope_type(self, item):
        try:
            return item.content_type.model_class().get_scope_type()
        except AttributeError:
            return None

    def get_scope_uuid(self, item):
        if item.scope:
            return item.scope.uuid.hex
        return item.details.get('scope_uuid')


class InvoiceSerializer(core_serializers.RestrictedSerializerMixin,
                        serializers.HyperlinkedModelSerializer):
    price = serializers.DecimalField(max_digits=15, decimal_places=7)
    tax = serializers.DecimalField(max_digits=15, decimal_places=7)
    total = serializers.DecimalField(max_digits=15, decimal_places=7)
    items = serializers.SerializerMethodField()
    issuer_details = serializers.SerializerMethodField()
    customer_details = serializers.SerializerMethodField()
    due_date = serializers.DateField()
    file = serializers.SerializerMethodField()

    class Meta:
        model = models.Invoice
        fields = (
            'url', 'uuid', 'number', 'customer', 'price', 'tax', 'total',
            'state', 'year', 'month', 'issuer_details', 'invoice_date', 'due_date',
            'customer', 'customer_details', 'items', 'file',
        )
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'customer': {'lookup_field': 'uuid'},
        }

    def get_issuer_details(self, invoice):
        return settings.WALDUR_INVOICES['ISSUER_DETAILS']

    def get_customer_details(self, invoice):
        return {
            'name': invoice.customer.name,
            'address': invoice.customer.address,
            'country': invoice.customer.country,
            'country_name': invoice.customer.get_country_display(),
            'email': invoice.customer.email,
            'postal': invoice.customer.postal,
            'phone_number': invoice.customer.phone_number,
            'bank_name': invoice.customer.bank_name,
            'bank_account': invoice.customer.bank_account,
        }

    def get_file(self, obj):
        if not obj.has_file():
            return None

        return reverse('invoice-pdf',
                       kwargs={'uuid': obj.uuid.hex},
                       request=self.context['request'])

    def get_items(self, invoice):
        items = utils.filter_invoice_items(invoice.items.all())
        serializer = GenericItemSerializer(items, many=True, context=self.context)
        return serializer.data


class InvoiceItemReportSerializer(serializers.ModelSerializer):
    invoice_number = serializers.ReadOnlyField(source='invoice.number')
    invoice_uuid = serializers.ReadOnlyField(source='invoice.uuid')
    invoice_year = serializers.ReadOnlyField(source='invoice.year')
    invoice_month = serializers.ReadOnlyField(source='invoice.month')
    invoice_date = serializers.ReadOnlyField(source='invoice.invoice_date')
    due_date = serializers.ReadOnlyField(source='invoice.due_date')
    customer_uuid = serializers.ReadOnlyField(source='invoice.customer.uuid')
    customer_name = serializers.ReadOnlyField(source='invoice.customer.name')

    class Meta:
        model = models.InvoiceItem
        fields = (
            'customer_uuid', 'customer_name',
            'project_uuid', 'project_name',
            'invoice_uuid', 'invoice_number',
            'invoice_year', 'invoice_month',
            'invoice_date', 'due_date',
            'invoice_price', 'invoice_tax', 'invoice_total',
            'name', 'article_code', 'product_code',
            'price', 'tax', 'total', 'unit_price', 'unit',
            'start', 'end',
        )
        decimal_fields = (
            'price', 'tax', 'total', 'unit_price',
            'invoice_price', 'invoice_tax', 'invoice_total'
        )
        decimal_fields_extra_kwargs = {
            'invoice_price': {
                'source': 'invoice.price',
            },
            'invoice_tax': {
                'source': 'invoice.tax',
            },
            'invoice_total': {
                'source': 'invoice.total',
            },
        }

    def build_field(self, field_name, info, model_class, nested_depth):
        if field_name in self.Meta.decimal_fields:
            field_class = serializers.DecimalField
            field_kwargs = dict(
                max_digits=20,
                decimal_places=2,
                coerce_to_string=True,
            )
            default_kwargs = self.Meta.decimal_fields_extra_kwargs.get(field_name)
            if default_kwargs:
                field_kwargs.update(default_kwargs)
            return field_class, field_kwargs

        return super(InvoiceItemReportSerializer, self).build_field(field_name, info, model_class, nested_depth)

    def get_extra_kwargs(self):
        extra_kwargs = super(InvoiceItemReportSerializer, self).get_extra_kwargs()
        extra_kwargs.update(settings.WALDUR_INVOICES['INVOICE_REPORTING']['SERIALIZER_EXTRA_KWARGS'])
        return extra_kwargs


# SAF is accounting soft from Estonia: www.sysdec.ee/safsaf.htm
class SAFReportSerializer(serializers.Serializer):
    DOKNR = serializers.ReadOnlyField(source='invoice.number')
    KUUPAEV = serializers.SerializerMethodField(method_name='get_last_day_of_month')
    VORMKUUP = serializers.SerializerMethodField(method_name='get_invoice_date')
    MAKSEAEG = serializers.SerializerMethodField(method_name='get_due_date')
    YKSUS = serializers.ReadOnlyField(source='invoice.customer.agreement_number')
    PARTNER = serializers.ReadOnlyField(source='invoice.customer.agreement_number')
    ARTIKKEL = serializers.ReadOnlyField(source='article_code')
    KOGUS = serializers.SerializerMethodField(method_name='get_quantity')
    SUMMA = serializers.SerializerMethodField(method_name='get_total')
    RMAKSUSUM = serializers.SerializerMethodField(method_name='get_tax')
    RMAKSULIPP = serializers.SerializerMethodField(method_name='get_vat')
    ARTPROJEKT = serializers.SerializerMethodField(method_name='get_project')
    ARTNIMI = serializers.ReadOnlyField(source='name')
    VALI = serializers.SerializerMethodField(method_name='get_vali_field')
    U_KONEDEARV = serializers.SerializerMethodField(method_name='get_empty_field')
    H_PERIOOD = serializers.SerializerMethodField(method_name='get_covered_period')

    class Meta:
        fields = ('DOKNR', 'KUUPAEV', 'VORMKUUP', 'MAKSEAEG', 'YKSUS', 'PARTNER',
                  'ARTIKKEL', 'KOGUS', 'SUMMA', 'RMAKSUSUM', 'RMAKSULIPP',
                  'ARTPROJEKT', 'ARTNIMI', 'VALI', 'U_KONEDEARV', 'H_PERIOOD')

    def format_date(self, date):
        if date:
            return date.strftime('%d.%m.%Y')
        return ''

    def get_first_day(self, invoice_item):
        year = invoice_item.invoice.year
        month = invoice_item.invoice.month
        return datetime.date(year=year, month=month, day=1)

    def get_last_day_of_month(self, invoice_item):
        first_day = self.get_first_day(invoice_item)
        last_day = core_utils.month_end(first_day)
        return self.format_date(last_day)

    def get_invoice_date(self, invoice_item):
        date = invoice_item.invoice.invoice_date
        return self.format_date(date)

    def get_due_date(self, invoice_item):
        date = invoice_item.invoice.due_date
        return self.format_date(date)

    def get_quantity(self, invoice_item):
        return invoice_item.get_factor(False)

    def get_total(self, invoice_item):
        return quantize_price(invoice_item.price)

    def get_tax(self, invoice_item):
        return quantize_price(invoice_item.tax)

    def get_project(self, invoice_item):
        return settings.WALDUR_INVOICES['INVOICE_REPORTING']['SAF_PARAMS']['ARTPROJEKT']

    def get_vat(self, invoice_item):
        return settings.WALDUR_INVOICES['INVOICE_REPORTING']['SAF_PARAMS']['RMAKSULIPP']

    def get_vali_field(self, invoice_item):
        return 'Record no %s' % invoice_item.invoice.number

    def get_empty_field(self, invoice_item):
        return ''

    def get_covered_period(self, invoice_item):
        first_day = self.get_first_day(invoice_item)
        last_day = core_utils.month_end(first_day)
        return '%s-%s' % (self.format_date(first_day), self.format_date(last_day))



# Generated by Django 2.2.7 on 2020-01-21 12:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace_checklist', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='question',
            name='solution',
            field=models.TextField(blank=True, null=True),
        ),
    ]



# Generated by Django 2.2.7 on 2019-12-27 13:58

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_rancher', '0008_node_annotations'),
    ]

    operations = [
        migrations.AddField(
            model_name='cluster',
            name='runtime_state',
            field=models.CharField(blank=True, max_length=255),
        ),
        migrations.AddField(
            model_name='node',
            name='runtime_state',
            field=models.CharField(blank=True, max_length=255),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-10-11 15:27
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0002_immutable_default_json'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='customer',
            options={'ordering': ('name',), 'verbose_name': 'organization'},
        ),
        migrations.AlterModelOptions(
            name='servicesettings',
            options={'ordering': ('name',), 'verbose_name': 'Service settings', 'verbose_name_plural': 'Service settings'},
        ),
    ]



from django.contrib.contenttypes.models import ContentType

from waldur_core.core import utils as core_utils
from waldur_mastermind.support import serializers as support_serializers


class OfferingCreateSerializer(support_serializers.OfferingCreateSerializer):
    def _get_issue_details(self, validated_data):
        issue_details = super(OfferingCreateSerializer, self)._get_issue_details(validated_data)
        order_item_serialized = self.context['request'].data.get('order_item')

        if order_item_serialized:
            order_item = core_utils.deserialize_instance(order_item_serialized)
            issue_details['resource_object_id'] = order_item.id
            issue_details['resource_content_type'] = ContentType.objects.get_for_model(order_item)
            issue_details['caller'] = order_item.order.created_by

        return issue_details



# Generated by Django 2.2.6 on 2019-11-13 12:30

from django.db import migrations, models
import django_fsm
import waldur_core.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_rancher', '0003_node_roles'),
    ]

    operations = [
        migrations.AddField(
            model_name='node',
            name='backend_id',
            field=models.CharField(blank=True, max_length=255),
        ),
        migrations.AddField(
            model_name='node',
            name='error_message',
            field=models.TextField(blank=True),
        ),
        migrations.AddField(
            model_name='node',
            name='name',
            field=models.CharField(default='', max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='node',
            name='state',
            field=django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5),
        ),
    ]



from django.conf import settings as django_settings
from django.db import migrations


def get_allocation_name(allocation):
    return get_account_name(django_settings.WALDUR_SLURM['ALLOCATION_PREFIX'], allocation)


def get_account_name(prefix, object_or_uuid):
    key = isinstance(object_or_uuid, str) and object_or_uuid or object_or_uuid.uuid.hex
    return '%s%s' % (prefix, key)


def fill_backend_id(apps, schema_editor):
    Allocation = apps.get_model('waldur_slurm', 'Allocation')
    for allocation in Allocation.objects.all():
        allocation.backend_id = get_allocation_name(allocation)
        allocation.save()


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_slurm', '0006_allocationusage_deposit_usage'),
    ]

    operations = [
        migrations.RunPython(fill_backend_id),
    ]



# Django test settings for Waldur Core.
from waldur_core.server.doc_settings import *  # noqa: F403

MEDIA_ROOT = '/tmp/'  # nosec

INSTALLED_APPS += (  # noqa: F405
    'waldur_core.quotas.tests',
    'waldur_core.structure.tests',
    'waldur_pid.tests',
)

ROOT_URLCONF = 'waldur_core.structure.tests.urls'

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.postgresql_psycopg2',
        'NAME': 'waldur',
    }
}

ALLOWED_HOSTS = ['localhost']
